"""
Useful things for plotting GeoClaw results.
"""

import numpy
import warnings

from clawpack.visclaw import colormaps

# Colormaps from geoclaw
# Color attributes, single instance per run
# Colors
black = [0.0,0.0,0.0]
white = [1.0,1.0,1.0]
green = [0.0,1.0,0.0];
dark_green = [0.1,0.4,0.0];
light_green = [0.8,1.0,0.5];
dark_blue = [0.2,0.2,0.7];
light_blue = [0.5,0.5,1.0];
blue = [0.0,0.0,1.0];
blue_green = [0.0,1.0,1.0]
red = [1.0,0.0,0.0]
tan = [0.9,0.8,0.2];
tan = [0.8,0.5,0.2];
brown = [0.9,0.8,0.2];
gray8 = [0.8,0.8,0.8];

transparent = [0.0, 0.0, 0.0,0.0]
blue_a = [0.0,0.0,1.0,1.0];
red_a = [1.0,0.0,0.0,1.0]
light_green_a = [0.0,1.0,1.0,1.0];
white_a = [1.0,1.0,1.0,1.0]

#Colors taken from Google Earth
biscay = '#203469'   # Dark blue
lightblue = '#4E6498'


# Colormaps
TSUNAMI_MAX_AMPLITUDE = 0.6
googleearth_lightblue = colormaps.make_colormap({-TSUNAMI_MAX_AMPLITUDE:blue_a,
                                                 0.0:lightblue,
                                                 TSUNAMI_MAX_AMPLITUDE:red_a})

googleearth_darkblue = colormaps.make_colormap({-TSUNAMI_MAX_AMPLITUDE:blue_a,
                                                0.0:biscay,
                                                TSUNAMI_MAX_AMPLITUDE:red_a})


googleearth_white = colormaps.make_colormap({-TSUNAMI_MAX_AMPLITUDE:blue_a,
                                                0.0:white_a,
                                                TSUNAMI_MAX_AMPLITUDE:red_a})


googleearth_transparent = colormaps.make_colormap({-1.0:light_green_a,
                                                   0.0:transparent,
                                                   1.0:red_a})

# Start transparency at 0.2
googleearth_flooding = colormaps.make_colormap({ -1:transparent,
                                                 -0.2:light_blue,
                                                 1.0:dark_blue})


tsunami_colormap = colormaps.make_colormap({-TSUNAMI_MAX_AMPLITUDE:blue,
                                            0.0:blue_green,
                                            TSUNAMI_MAX_AMPLITUDE:red})

land1_colormap = colormaps.make_colormap({0.0:dark_green,
                                          1000.0:green,
                                          2000.0:light_green,
                                          4000.0:tan})

land2_colormap = colormaps.make_colormap({0:dark_green,
                                          50:green,
                                          100:light_green,
                                          200:tan})

water_land_colormap = colormaps.make_colormap({-1000:dark_blue,
                                               -500:blue,
                                               0:light_blue,
                                               .1:tan,
                                               5:tan,
                                               6:dark_green,
                                               1000:green,
                                               2000:light_green,
                                               4000:tan})

bathy1_colormap = colormaps.make_colormap({-1000:brown,
                                           0:tan,
                                           .1:dark_green,
                                           1000:green,
                                           2000:light_green})

bathy2_colormap = colormaps.make_colormap({-1000:brown,
                                           -100:tan,
                                           0:dark_green,
                                           .1:dark_green,
                                           1000:green,
                                           2000:light_green})

bathy3_colormap = colormaps.make_colormap({-1:[0.3,0.2,0.1],
                                           -0.01:[0.95,0.9,0.7],
                                           .01:[.5,.7,0],
                                           1:[.2,.5,.2]})

seafloor_colormap = colormaps.make_colormap({-1:[0.3,0.2,0.1],
                                              0:[0.95,0.9,0.7]})

land_colormap = colormaps.make_colormap({ 0:[0.95,0.9,0.7],
                                          1:[.2,.5,.2]})



colormaps_list = {"tsunami":tsunami_colormap,"land1":land1_colormap,
             "land2":land2_colormap,"water_land":water_land_colormap,
                  "bathy1":bathy1_colormap,"bathy2":bathy2_colormap}

def plot_colormaps():
    r"""Plots all colormaps avaiable or the ones specified"""

    import numpy as np
    import matplotlib.pyplot as plt

    a = np.linspace(0, 1, 256).reshape(1,-1)
    a = np.vstack((a,a))

    nmaps = len(colormaps_list) + 1

    fig = plt.figure(figsize=(5,10))
    fig.subplots_adjust(top=0.99, bottom=0.01, left=0.2, right=0.99)

    for i,name in enumerate(colormaps_list):
        ax = plt.subplot(nmaps,1,i+1)
        plt.axis("off")
        plt.imshow(a, aspect='auto', cmap=colormaps_list[name], origin='lower')
        pos = list(ax.get_position().bounds)
        fig.text(pos[0] - 0.01, pos[1], name, fontsize=10, horizontalalignment='right')

    #plt.show()

land_colors = colormaps.make_colormap({0:[.5,.7,0], 1:[.2,.5,.2]})
# water_colors = colormaps.make_colormap({-1.:'r', 0.:[0, .8, .8], 1.: 'b'})
# land_colors = land2_colormap
water_colors = tsunami_colormap

# Plotting functions

# The drytol parameter is used in masking land and water and
# affects what color map is used for cells with small water depth h.
# The best value to use often depends on the application and can
# be set for an application by setting current_data.user['dry_tolerance'] in
# a beforeframe function, for example.  If it's not set by the user,
# the following default value is used (in meters):

drytol_default = 1.e-3

def topo(current_data):
   """
   Return topography = eta - h.
   Surface eta is assumed to be output as 4th column of fort.q files.
   """
   q = current_data.q
   h = q[0,...]
   eta = q[-1,...]
   topo = eta - h
   return topo


def land(current_data):
   """
   Return a masked array containing the surface elevation only in dry cells.
   """
   drytol = current_data.user.get('dry_tolerance', drytol_default)
   q = current_data.q
   h = q[0,...]
   eta = q[-1,...]
   land = numpy.ma.masked_where(h>drytol, eta)
   return land


def water(current_data):
   """Deprecated: use surface instead."""
   raise DeprecationWarning("Deprecated function, use surface instead.")
   drytol = current_data.user.get('dry_tolerance', drytol_default)
   q = current_data.q
   h = q[0,...]
   eta = q[-1,...]
   water = numpy.ma.masked_where(h<=drytol, eta)
   return water


def depth(current_data):
   """
   Return a masked array containing the depth of fluid only in wet cells.
   """
   drytol = current_data.user.get('dry_tolerance', drytol_default)
   q = current_data.q
   h = q[0,...]
   depth = numpy.ma.masked_where(h<=drytol, h)
   try:
       # Use mask covering coarse regions if it's set:
       m = current_data.mask_coarse
       depth = numpy.ma.masked_where(m, depth)
   except:
       pass

   return depth


def surface(current_data):
    """
    Return a masked array containing the surface elevation only in wet cells.
    Surface is eta = h+topo, assumed to be output as 4th column of fort.q
    files.
    """
    drytol = current_data.user.get('dry_tolerance', drytol_default)
    q = current_data.q
    h = q[0,...]
    eta = q[-1,...]

    water = numpy.ma.masked_where(h <= drytol,eta)

    try:
        # Use mask covering coarse regions if it's set:
        m = current_data.mask_coarse
        water = numpy.ma.masked_where(m, water)
    except:
        pass

    return water


def surface_or_depth(current_data):
    """
    Return a masked array containing the surface elevation where the topo is
    below sea level or the water depth where the topo is above sea level.
    Mask out dry cells.  Assumes sea level is at topo=0.
    Surface is eta = h+topo, assumed to be output as 4th column of fort.q
    files.
    """

    drytol = current_data.user.get('dry_tolerance', drytol_default)
    q = current_data.q
    h = q[0,...]
    eta = q[-1,...]
    topo = eta - h

    # With this version, the land was plotted as white in png files for KML.
    # surface = ma.masked_where(h <= drytol, eta)
    # depth = ma.masked_where(h <= drytol, h)
    # surface_or_depth = where(topo<0, surface, depth)

    # With this version, the land is transparent.
    surface_or_depth = numpy.ma.masked_where(h <= drytol,
                                             numpy.where(topo<0, eta, h))

    try:
        # Use mask covering coarse regions if it's set:
        m = current_data.mask_coarse
        surface_or_depth = numpy.ma.masked_where(m, surface_or_depth)
    except:
        pass

    return surface_or_depth


def u_velocity(current_data):
    """
    Return a masked array containing the u velocity (x-component)
    Mask out dry cells.  
    """

    drytol = current_data.user.get('dry_tolerance', drytol_default)
    q = current_data.q
    h = q[0,...]
    hu = q[1,...]
    h_wet = numpy.ma.masked_where(h<=drytol, h)
    u_wet = hu / h_wet

    try:
        # Use mask covering coarse regions if it's set:
        m = current_data.mask_coarse
        u_wet = numpy.ma.masked_where(m, u_wet)
    except:
        pass

    return u_wet


def v_velocity(current_data):
    """
    Return a masked array containing the v velocity (y-component)
    Mask out dry cells.  
    """

    drytol = current_data.user.get('dry_tolerance', drytol_default)
    q = current_data.q
    h = q[0,...]
    hv = q[2,...]
    h_wet = numpy.ma.masked_where(h<=drytol, h)
    v_wet = hv / h_wet

    try:
        # Use mask covering coarse regions if it's set:
        m = current_data.mask_coarse
        v_wet = numpy.ma.masked_where(m, v_wet)
    except:
        pass

    return v_wet

def speed(current_data):
    """
    Return a masked array containing the speed.
    Mask out dry cells.  
    """

    drytol = current_data.user.get('dry_tolerance', drytol_default)
    q = current_data.q
    h = q[0,...]
    hu = q[1,...]
    hv = q[2,...]
    h_wet = numpy.ma.masked_where(h<=drytol, h)
    u_wet = hu / h_wet
    v_wet = hv / h_wet
    speed = numpy.sqrt(u_wet**2 + v_wet**2)

    try:
        # Use mask covering coarse regions if it's set:
        m = current_data.mask_coarse
        speed = numpy.ma.masked_where(m, speed)
    except:
        pass

    return speed



def kml_build_colorbar(cb_filename, cmap, cmin, cmax):

    """
    This version is deprecated, please use 
        clawpack.geoclaw.kmltools.kml_build_colorbar
    """
    import matplotlib.pyplot as plt
    import matplotlib as mpl
    import warnings

    msg = "geoplot.kml_build_colorbar is deprecated, " \
        + "instead use clawpack.geoclaw.kmltools.kml_build_colorbar"
    warnings.warn(msg)

    fig = plt.figure(figsize=(0.8,3))
    ax1 = fig.add_axes([0.1, 0.075, 0.25, 0.85])
    tick = ax1.yaxis.get_major_ticks()
    plt.tick_params(axis='y', which='major', labelsize=8)

    norm = mpl.colors.Normalize(vmin=cmin,vmax=cmax)

    cb1 = mpl.colorbar.ColorbarBase(ax1,cmap=cmap,
                                    norm=norm,
                                    orientation='vertical')
    # This is called from plotpages, in <plotdir>.
    plt.savefig(cb_filename,Transparent=True)


# Some discrete color maps useful for contourf plots of fgmax results:

def discrete_cmap_1(clines):
    """
    Construct a discrete color map for the regions between the contour lines
    given in clines. Colors go from turqouise through yellow to red.
    Good for zeta.
    """
    from numpy import floor, linspace, hstack, ones, zeros
    nlines = len(clines)
    n1 = int(floor((nlines-1)/2.))
    n2 = nlines - 1 - n1
    Green = hstack([linspace(1,1,n1),linspace(1,0,n2)])
    Red = hstack([linspace(0,0.8,n1), ones(n2)])
    Blue = hstack([linspace(1,0.2,n1), zeros(n2)])
    colors = list(zip(Red,Green,Blue))
    return colors

def discrete_cmap_2(clines):
    """
    Construct a discrete color map for the regions between the contour lines
    given in clines.  Colors go from red to turquoise.
    Good for arrival times.
    """
    from numpy import floor, linspace, hstack, ones, zeros, flipud
    nlines = len(clines)
    n1 = int(floor((nlines-1)/2.))
    n2 = nlines - 1 - n1
    Green = flipud(hstack([linspace(1,1,n1),linspace(1,0,n2)]))
    Red = flipud(hstack([linspace(0,0.8,n1), ones(n2)]))
    Blue = flipud(hstack([linspace(1,0.2,n1), zeros(n2)]))
    colors = list(zip(Red,Green,Blue))
    return colors



#!/usr/bin/env python
# encoding: utf-8
r"""
Plotting routines for multilayer simulations with GeoClaw

:Authors:
    Kyle Mandli (2011-2-07) Initial version
"""

import numpy
import matplotlib.pyplot as plt

from clawpack.visclaw import colormaps, geoplot, gaugetools
from clawpack.visclaw.data import ClawPlotData
import clawpack.geoclaw.data
import os
plotdata = ClawPlotData()

# Definition of colormpas
surface_cmap = plt.get_cmap("bwr")
speed_cmap = plt.get_cmap('PuBu')
land_cmap = geoplot.land_colors
velocity_cmap = plt.get_cmap('PiYG')
vorticity_cmap = plt.get_cmap('PRGn')

# ========================================================================
#  Data extraction routines
#  This array contains the location of the start of each layer's data.  To
#  access a layer's depth for instance you want to use
#    q[layer_index[0] + 1, :, :]
#  Momenta in the x and y directions are `+ 2` and `+ 3` respectively.
#
#  Note also that all layer indices are 0-indexed
# ======================================================================== 
layer_index = [0, 3]
eta_index = 6


def layered_land(surface, DRY_TOL=10**-3):

    def land(cd):
        """
        Return a masked array containing the surface elevation only in dry cells.
        """
        q = cd.q
        h = q[(surface-1)*3,:,:]
        eta = q[surface+5,:,:]
        land = numpy.ma.masked_where(h>DRY_TOL, eta)
        return land
    
    return land

def extract_eta(h,eta,DRY_TOL=10**-3):
    masked_eta = numpy.ma.masked_where(numpy.abs(h) < DRY_TOL, eta)
    return masked_eta


def eta(cd, layer):
    return extract_eta(cd.q[layer_index[layer], :, :],
                       cd.q[eta_index + layer, :, :])


def eta1(cd):
    return eta(cd, 0)


def eta2(cd):
    return eta(cd, 1)


def eta_elevation(surface, DRY_TOL=10**-3):

    def eta_func(cd):
        ml_data = clawpack.geoclaw.data.MultilayerData()
        ml_data.read(os.path.join(plotdata.outdir,'multilayer.data'))

        h1 = cd.q[0,:,:]
        h2 = cd.q[3,:,:]
        b = cd.q[6,:,:] - h1 - h2
        if surface == 1: 
            h = h1
            eta = eta1(cd)
            eta_init = ml_data.eta[0]
        elif surface == 2:
            h = h2
            eta = eta2(cd)
            eta_init = ml_data.eta[1]
        return numpy.ma.masked_where(b < eta_init, eta, numpy.ma.masked_where(numpy.abs(h) < DRY_TOL, h))

    return eta_func

def initially_wet(surface, DRY_TOL=10**-3):

    def eta_func(cd):
        ml_data = clawpack.geoclaw.data.MultilayerData()
        ml_data.read(os.path.join(plotdata.outdir,'multilayer.data'))

        h1 = cd.q[0,:,:]
        h2 = cd.q[3,:,:]
        b = cd.q[6,:,:] - h1 - h2
        if surface == 1: 
            h = h1
            eta = eta1(cd)
            eta_init = ml_data.eta[0]
        elif surface == 2:
            h = h2
            eta = eta2(cd)
            eta_init = ml_data.eta[1]
        mask = numpy.logical_or(b>eta_init, numpy.abs(h)<DRY_TOL)
        return numpy.ma.masked_where(mask, eta)

    return eta_func

def inundated(surface, DRY_TOL=10**-3):

    def h_func(cd):
        ml_data = clawpack.geoclaw.data.MultilayerData()
        ml_data.read(os.path.join(plotdata.outdir,'multilayer.data'))

        h1 = cd.q[0,:,:]
        h2 = cd.q[3,:,:]
        b = cd.q[6,:,:] - h1 - h2
        if surface == 1: 
            h = h1
            eta = eta1(cd)
            eta_init = ml_data.eta[0]
        elif surface == 2:
            h = h2
            eta = eta2(cd)
            eta_init = ml_data.eta[1]
        mask = numpy.logical_or(b<eta_init, numpy.abs(h)<DRY_TOL)
        return numpy.ma.masked_where(mask, h)

    return h_func

# def h1(cd):
#     # return cd.q[:,:,6]
#     return extract_h(cd.q[0,:,:],cd.q[6,:,:])
    
# def h2(cd):
#     # return cd.q[:,:,7]
#     return extract_h(cd.q[3,:,:],cd.q[7,:,:])

def b(current_data):
    h1 = current_data.q[layer_index[0], :, :]
    h2 = current_data.q[layer_index[1], :, :]

    return eta1(current_data) - h1 - h2


def extract_velocity(h, hu, DRY_TOL=10**-8):
    u = numpy.ones(hu.shape) * numpy.nan
    index = numpy.nonzero((numpy.abs(h) > DRY_TOL) * (h != numpy.nan))
    u[index[0], index[1]] = hu[index[0], index[1]] / h[index[0], index[1]]
    return u


def water_u(cd, direction, layer):
    return extract_velocity(cd.q[layer_index[layer], :, :],
                            cd.q[layer_index[layer] + direction + 1, :, :])


def water_u1(cd):
    return water_u(cd, 0, 0)


def water_u2(cd):
    return water_u(cd, 0, 1)


def water_v1(cd):
    return water_u(cd, 1, 0)


def water_v2(cd):
    return water_u(cd, 1, 1)


def water_speed1(current_data):
    u = water_u1(current_data)
    v = water_v1(current_data)

    return numpy.sqrt(u**2 + v**2)


def water_speed2(current_data):
    u = water_u2(current_data)
    v = water_v2(current_data)

    return numpy.sqrt(u**2 + v**2)


def water_speed_depth_ave(current_data):
    h1 = current_data.q[layer_index[0], :, :]
    h2 = current_data.q[layer_index[1], :, :]
    u1 = water_speed1(current_data)
    u2 = water_speed1(current_data)

    return (h1 * u1 + h2 * u2) / (h1 + h2)


# ========================================================================
#  Plot items
# ========================================================================
def add_inundation(plotaxes,surface,bounds=None,plot_type='pcolor'):
    if plot_type == 'pcolor' or plot_type == 'imshow':
        plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
        plotitem.plot_var = inundated(surface)        
        if bounds is not None:                
            plotitem.pcolor_cmin = bounds[0]
            plotitem.pcolor_cmax = bounds[1]
        plotitem.pcolor_cmap = colormaps.make_colormap({1.0:'c',0.0:'w'})
        plotitem.add_colorbar = True
        plotitem.amr_celledges_show = [0,0,0]
        plotitem.amr_patchedges_show = [0,0,0]

def add_surface_elevation(plotaxes,surface,bounds=None,plot_type='pcolor'):
    if plot_type == 'pcolor' or plot_type == 'imshow':
        plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
        # plotitem.plot_var = geoplot.surface
        # plotitem.plot_var = eta_elevation(surface)
        plotitem.plot_var = initially_wet(surface)        

        # if surface == 1:
        #     plotitem.plot_var = eta1
        # elif surface == 2:
        #     plotitem.plot_var = eta2
        if bounds is not None:                
            plotitem.pcolor_cmin = bounds[0]
            plotitem.pcolor_cmax = bounds[1]
        plotitem.pcolor_cmap = surface_cmap
        plotitem.add_colorbar = True
        plotitem.amr_celledges_show = [0] * 10
        plotitem.amr_patchedges_show = [0] * 10

    elif plot_type == 'contour':
        plotitem = plotaxes.new_plotitem(plot_type='2d_contour')
        plotitem.plot_var = surface + 5
        if bounds is not None:
            plotitem.contour_levels = bounds
        plotitem.amr_contour_show = [1] * 10
        plotitem.amr_celledges_show = [0] * 10
        plotitem.amr_patchedges_show = [1] * 10
        plotitem.amr_contour_colors = 'k'

    else:
        raise NotImplementedError("Plot type %s not implemented" % plot_type)


def add_layer_depth(plotaxes, layer, bounds=None, plot_type='pcolor'):

    if plot_type == 'pcolor' or plot_type == 'imshow':
        plotitem = plotaxes.new_plotitem(plot_type='2d_imshow')
        if layer == 1:
            plotitem.plot_var = 0
        elif layer == 2:
            plotitem.plot_var = 3
        if bounds is not None:
            plotitem.imshow_cmin = bounds[0]
            plotitem.imshow_cmax = bounds[1]
        plotitem.imshow_cmap = surface_cmap
        plotitem.add_colorbar = True
        plotitem.amr_celledges_show = [0] * 10
        plotitem.amr_patchedges_show = [1] * 10


def add_speed(plotaxes, layer, bounds=None, plot_type='pcolor'):

    if plot_type == 'pcolor' or plot_type == 'imshow':
        plotitem = plotaxes.new_plotitem(plot_type='2d_imshow')
        if layer == 1:
            plotitem.plot_var = water_speed1
        elif layer == 2:
            plotitem.plot_var = water_speed2
        plotitem.imshow_cmap = speed_cmap
        if bounds is not None:
            plotitem.imshow_cmin = bounds[0]
            plotitem.imshow_cmax = bounds[1]
        plotitem.add_colorbar = True
        plotitem.amr_celledges_show = [0] * 10
        plotitem.amr_patchedges_show = [1] * 10


def add_x_velocity(plotaxes, layer, plot_type='pcolor', bounds=None):

    if plot_type == 'pcolor' or plot_type == 'imshow':
        plotitem = plotaxes.new_plotitem(plot_type='2d_imshow')
        if layer == 1:
            plotitem.plot_var = water_u1
        if layer == 2:
            plotitem.plot_var = water_u2
        if bounds is not None:
            plotitem.imshow_cmin = bounds[0]
            plotitem.imshow_cmax = bounds[1]
        plotitem.add_colorbar = True
        plotitem.imshow_cmap = velocity_cmap
        plotitem.amr_celledges_show = [0] * 10
        plotitem.amr_patchedges_show = [1] * 10


def add_y_velocity(plotaxes, layer, plot_type='pcolor', bounds=None):

    if plot_type == 'pcolor' or plot_type == 'imshow':
        plotitem = plotaxes.new_plotitem(plot_type='2d_imshow')
        if layer == 1:
            plotitem.plot_var = water_v1
        if layer == 2:
            plotitem.plot_var = water_v2
        if bounds is not None:
            plotitem.imshow_cmin = bounds[0]
            plotitem.imshow_cmax = bounds[1]
        plotitem.imshow_cmap = velocity_cmap
        plotitem.add_colorbar = True
        plotitem.amr_celledges_show = [0] * 10
        plotitem.amr_patchedges_show = [1] * 10


# Land
def add_land(plotaxes, surface, plot_type='pcolor', bounds=[-10, 10]):
    r"""Add plot item for land"""

    if plot_type == 'pcolor':
        plotitem = plotaxes.new_plotitem(plot_type='2d_imshow')
        plotitem.show = True
        plotitem.plot_var = layered_land(surface)
        plotitem.imshow_cmap = land_cmap
        plotitem.imshow_cmin = bounds[0]
        plotitem.imshow_cmax = bounds[1]
        plotitem.add_colorbar = False
        plotitem.amr_celledges_show = [0] * 10
        plotitem.amr_patchedges_show = [1] * 10

    elif plot_type == 'contour':
        plotitem = plotaxes.new_plotitem(plot_type='2d_contour')
        plotitem.plot_var = geoplot.land
        plotitem.contour_nlevels = 40
        plotitem.contour_min = bounds[0]
        plotitem.contour_max = bounds[1]
        plotitem.amr_contour_colors = ['g']  # color on each level
        plotitem.amr_patch_bgcolor = ['#ffeeee', '#eeeeff', '#eeffee']
        plotitem.celledges_show = 0
        plotitem.patchedges_show = 0
        plotitem.show = True

    else:
        raise NotImplementedError("Plot type %s not implemented" % plot_type)


def add_combined_profile_plot(plot_data, slice_value, direction='x',
                              figno=120):
    def slice_index(cd):
        if direction == 'x':
            if cd.grid.y.lower < slice_value < cd.grid.y.upper:
                return int((slice_value - cd.grid.y.lower) / cd.dy - 0.5)
            else:
                return None
        elif direction == 'y':
            if cd.grid.x.lower < slice_value < cd.grid.x.upper:
                return int((slice_value - cd.grid.x.lower) / cd.dx - 0.5)
            else:
                return None

    def bathy_profile(current_data):
        index = slice_index(current_data)
        if direction == 'x':
            if index:
                return current_data.x[:, index], b(current_data)[:, index]
            else:
                return None
        elif direction == 'y':
            if index:
                return current_data.y[index, :], b(current_data)[index, :]
            else:
                return None

    def lower_surface(current_data):
        index = slice_index(current_data)
        if direction == 'x':
            if index:
                return current_data.x[:, index], eta2(current_data)[:, index]
            else:
                return None
        elif direction == 'y':
            if index:
                return current_data.y[index, :], eta2(current_data)[index, :]
            else:
                return None

    def upper_surface(current_data):
        index = slice_index(current_data)
        if direction == 'x':
            if index:
                return current_data.x[:, index], eta1(current_data)[:, index]
            else:
                return None
        elif direction == 'y':
            if index:
                return current_data.y[index, :], eta1(current_data)[index, :]
            else:
                return None

    # Surfaces
    plotfigure = plotdata.new_plotfigure(name='combined_surface_%s' % figno,
                                         figno=figno)
    plotfigure.show = True
    plotfigure.kwargs = {'figsize': (6, 6)}

    # Top surface
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.axescmd = 'subplot(2,1,1)'
    plotaxes.title = 'Surfaces Profile %s at %s' % (direction, slice_value)
    if multilayer_data.init_type == 2:
        plotaxes.xlimits = xlimits
    elif multilayer_data.init_type == 6:
        plotaxes.xlimits = ylimits

    plotaxes.ylimits = top_surf_zoomed

    def top_surf_afteraxes(cd):
        axes = plt.gca()
        axes.set_xlabel('')
        locs, labels = plt.xticks()
        labels = ['' for i in range(len(locs))]
        plt.xticks(locs, labels)
        axes.plot([multilayer_data.bathy_location,
                  multilayer_data.bathy_location], top_surf_zoomed, '--k')
        axes.set_ylabel('m')

    plotaxes.afteraxes = top_surf_afteraxes
    plotitem = plotaxes.new_plotitem(plot_type="1d_from_2d_data")
    plotitem.map_2d_to_1d = upper_surface
    plotitem.amr_plotstyle = ['-', '+', 'x']
    plotitem.show = True

    # Internal surface
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.axescmd = 'subplot(2,1,2)'
    plotaxes.title = ''
    if multilayer_data.init_type == 2:
        plotaxes.xlimits = xlimits
    elif multilayer_data.init_type == 6:
        plotaxes.xlimits = ylimits
    plotaxes.ylimits = bottom_surf_zoomed

    def internal_surf_afteraxes(cd):
        axes = plt.gca()
        axes.set_title('')
        axes.set_ylabel('m')
        axes.subplots_adjust(hspace=0.05)
        axes.plot([multilayer_data.bathy_location,
                  multilayer_data.bathy_location], bottom_surf_zoomed, '--k')
    plotaxes.afteraxes = internal_surf_afteraxes
    plotitem = plotaxes.new_plotitem(plot_type='1d_from_2d_data')
    plotitem.map_2d_to_1d = lower_surface
    plotitem.amr_plotstyle = ['-', '+', 'x']
    plotitem.color = 'k'
    plotitem.show = True


def add_velocities_profile_plot(plot_data, slice_value, direction='x',
                                figno=130):

    def slice_index(cd):
        if cd.grid.y.lower < slice_value < cd.grid.y.upper:
            return int((slice_value - cd.grid.y.lower) / cd.dy - 0.5)
        else:
            return None

    def upper_surface(current_data):
        index = slice_index(current_data)
        if index:
            return current_data.x[:, index], eta1(current_data)[:, index]
        else:
            return None

    def top_speed(current_data):
        index = slice_index(current_data)
        if index:
            return current_data.x[:, index], water_u1(current_data)[:, index]
        else:
            return None, None

    def bottom_speed(current_data):
        index = slice_index(current_data)
        if index:
            return current_data.x[:, index], water_u2(current_data)[:, index]
        else:
            return None, None

    # Velocities
    plotfigure = plotdata.new_plotfigure(name='combined_velocities_%s'
                                         % figno, figno=figno)
    plotfigure.show = True

    # Top surface
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'Velocities Profile %s at %s' % (direction, slice_value)
    if multilayer_data.init_type == 2:
        plotaxes.xlimits = xlimits
    elif multilayer_data.init_type == 6:
        plotaxes.xlimits = ylimits
    plotaxes.ylimits = velocities_zoomed

    def velocity_afteraxes(cd):
        axes = plt.gca()
        axes.set_xlabel('')
        locs, labels = plt.xticks()
        labels = ['' for i in range(len(locs))]
        plt.xticks(locs, labels)
        axes.plot([multilayer_data.bathy_location,
                  multilayer_data.bathy_location], velocities_zoomed, '--k')
        axes.set_ylabel('m/s')
    plotaxes.afteraxes = velocity_afteraxes

    plotitem = plotaxes.new_plotitem(plot_type="1d_from_2d_data")
    plotitem.map_2d_to_1d = top_speed
    plotitem.amr_plotstyle = ['-', '+', 'x']
    plotitem.show = True

    plotitem = plotaxes.new_plotitem(plot_type='1d_from_2d_data')
    plotitem.map_2d_to_1d = bottom_speed
    plotitem.amr_plotstyle = ['-', '+', 'x']
    plotitem.color = 'k'
    plotitem.show = True




def add_cross_section(plotaxes, surface):
    r""" Add cross section view of surface"""
    if surface == 1: 
        plot_eta = eta1
        clr = 'c'
        sty = 'x'
    if surface == 2: 
        plot_eta = eta2
        clr = 'b'
        sty = '+'

    def xsec(current_data):
        # Return x value and surface eta at this point, along y=0
        from pylab import where,ravel
        x = current_data.x
        y = ravel(current_data.y)
        dy = current_data.dy

        ij = where((y <= dy/2.) & (y > -dy/2.))
        x_slice = ravel(x)[ij]
        eta_slice = ravel(plot_eta(current_data))[ij]
        return x_slice, eta_slice

    plotitem = plotaxes.new_plotitem(plot_type='1d_from_2d_data')
    plotitem.map_2d_to_1d = xsec
    plotitem.color = clr
    plotitem.plotstyle = sty

    plotitem.show = True

def add_land_cross_section(plotaxes):
    r""" Add cross section view of topo"""

    def plot_topo_xsec(current_data):
        from pylab import where,ravel
        x = current_data.x
        y = ravel(current_data.y)
        dy = current_data.dy

        ij = where((y <= dy/2.) & (y > -dy/2.))
        x_slice = ravel(x)[ij]
        b_slice = ravel(b(current_data))[ij]
        return x_slice, b_slice

    plotitem = plotaxes.new_plotitem(plot_type='1d_from_2d_data')
    plotitem.map_2d_to_1d = plot_topo_xsec
    plotitem.color = 'g'

    plotitem.show = True





#!/usr/bin/env python
# encoding: utf-8
r"""
Solitary wave formation in periodic nonlinear elastic media
===========================================================

Solve a one-dimensional nonlinear elasticity system:

.. math::
    \epsilon_t + u_x & = 0 \\
    (\rho(x) u)_t + \sigma(\epsilon,x)_x & = 0.

Here :math:`\epsilon` is the strain, :math:`\sigma` is the stress, 
u is the velocity, and :math:`\rho(x)` is the density.  
We take the stress-strain relation :math:`\sigma = e^{K(x)\epsilon}-1`;
:math:`K(x)` is the linearized bulk modulus.
Note that the density and bulk modulus may depend explicitly on x.

The problem solved here is based on [LeVYon03]_.  An initial hump
evolves into two trains of solitary waves.

"""
import numpy as np

def qinit(state,ic=2,a2=1.0,xupper=600.):
    x = state.grid.x.centers
    
    if ic==1: #Zero ic
        state.q[:,:] = 0.
    elif ic==2:
        # Gaussian
        sigma = a2*np.exp(-((x-xupper/2.)/10.)**2.)
        state.q[0,:] = np.log(sigma+1.)/state.aux[1,:]
        state.q[1,:] = 0.


def setaux(x,rhoB=4,KB=4,rhoA=1,KA=1,alpha=0.5,xlower=0.,xupper=600.,bc=2):
    aux = np.empty([3,len(x)],order='F')
    xfrac = x-np.floor(x)
    #Density:
    aux[0,:] = rhoA*(xfrac<alpha)+rhoB*(xfrac>=alpha)
    #Bulk modulus:
    aux[1,:] = KA  *(xfrac<alpha)+KB  *(xfrac>=alpha)
    aux[2,:] = 0. # not used
    return aux

    
def b4step(solver,state):
    #Reverse velocity at trtime
    #Note that trtime should be an output point
    if state.t>=state.problem_data['trtime']-1.e-10 and not state.problem_data['trdone']:
        #print 'Time reversing'
        state.q[1,:]=-state.q[1,:]
        state.q=state.q
        state.problem_data['trdone']=True
        if state.t>state.problem_data['trtime']:
            print('WARNING: trtime is '+str(state.problem_data['trtime'])+\
                ' but velocities reversed at time '+str(state.t))
    #Change to periodic BCs after initial pulse 
    if state.t>5*state.problem_data['tw1'] and solver.bc_lower[0]==0:
        solver.bc_lower[0]=2
        solver.bc_upper[0]=2
        solver.aux_bc_lower[0]=2
        solver.aux_bc_upper[0]=2


def zero_bc(state,dim,t,qbc,auxbc,num_ghost):
    """Set everything to zero"""
    if dim.on_upper_boundary:
        qbc[:,-num_ghost:]=0.

def moving_wall_bc(state,dim,t,qbc,auxbc,num_ghost):
    """Initial pulse generated at left boundary by prescribed motion"""
    if dim.on_lower_boundary:
        qbc[0,:num_ghost]=qbc[0,num_ghost] 
        t=state.t; t1=state.problem_data['t1']; tw1=state.problem_data['tw1']
        a1=state.problem_data['a1'];
        t0 = (t-t1)/tw1
        if abs(t0)<=1.: vwall = -a1*(1.+np.cos(t0*np.pi))
        else: vwall=0.
        for ibc in range(num_ghost-1):
            qbc[1,num_ghost-ibc-1] = 2*vwall*state.aux[1,ibc] - qbc[1,num_ghost+ibc]



def setup(use_petsc=0,kernel_language='Fortran',solver_type='classic',outdir='./_output',
          tfinal=500., num_output_times = 20):
    from clawpack import riemann

    if use_petsc:
        import clawpack.petclaw as pyclaw
    else:
        from clawpack import pyclaw

    if kernel_language=='Python':
        rs = riemann.nonlinear_elasticity_1D_py.nonlinear_elasticity_1D
    elif kernel_language=='Fortran':
        rs = riemann.nonlinear_elasticity_fwave_1D

    if solver_type=='sharpclaw':
        solver = pyclaw.SharpClawSolver1D(rs)
        solver.char_decomp=0
    else:
        solver = pyclaw.ClawSolver1D(rs)

    solver.kernel_language = kernel_language

    solver.bc_lower[0] = pyclaw.BC.custom
    solver.bc_upper[0] = pyclaw.BC.extrap

    #Use the same BCs for the aux array
    solver.aux_bc_lower[0] = pyclaw.BC.extrap
    solver.aux_bc_upper[0] = pyclaw.BC.extrap

    xlower=0.0; xupper=300.0
    cells_per_layer=12; mx=int(round(xupper-xlower))*cells_per_layer
    x = pyclaw.Dimension(xlower,xupper,mx,name='x')
    domain = pyclaw.Domain(x)
    state = pyclaw.State(domain,solver.num_eqn,num_aux=3)
    state.problem_data['stress_relation'] = 'exponential'

    #Set global parameters
    alpha = 0.5
    KA    = 1.0
    KB    = 4.0
    rhoA  = 1.0
    rhoB  = 4.0
    state.problem_data = {}
    state.problem_data['t1']    = 10.0
    state.problem_data['tw1']   = 10.0
    state.problem_data['a1']    = 0.1
    state.problem_data['alpha'] = alpha
    state.problem_data['KA'] = KA
    state.problem_data['KB'] = KB
    state.problem_data['rhoA'] = rhoA
    state.problem_data['rhoB'] = rhoB
    state.problem_data['trtime'] = 999999999.0
    state.problem_data['trdone'] = False

    #Initialize q and aux
    xc=state.grid.x.centers
    state.aux[:,:] = setaux(xc,rhoB,KB,rhoA,KA,alpha,xlower=xlower,xupper=xupper)
    qinit(state,ic=1,a2=1.0,xupper=xupper)

    solver.max_steps = 5000000
    solver.fwave = True 
    solver.before_step = b4step 
    solver.user_bc_lower=moving_wall_bc
    solver.user_bc_upper=zero_bc

    claw = pyclaw.Controller()
    claw.num_output_times = num_output_times
    claw.outdir = outdir
    claw.output_style = 1
    if outdir is None:
        claw.output_format = None
    claw.tfinal = tfinal
    claw.solution = pyclaw.Solution(state,domain)
    claw.solver = solver
    claw.setplot = setplot
    claw.keep_copy = True

    return claw


#--------------------------
def setplot(plotdata):
#--------------------------
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of visclaw.data.ClawPlotData.
    Output: a modified version of plotdata.
    """ 
    plotdata.clearfigures()  # clear any old figures,axes,items data

    # Figure for q[0]
    plotfigure = plotdata.new_plotfigure(name='Stress', figno=1)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'Stress'
    plotaxes.ylimits = [-0.1,1.0]

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = stress
    plotitem.kwargs = {'linewidth':2}
    
    # Figure for q[1]
    plotfigure = plotdata.new_plotfigure(name='Velocity', figno=2)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = [-.5,0.1]
    plotaxes.title = 'Velocity'

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = velocity
    plotitem.kwargs = {'linewidth':2}
    
    return plotdata

 
def velocity(current_data):
    """Compute velocity from strain and momentum"""
    aux=setaux(current_data.x,rhoB=4,KB=4)
    velocity = current_data.q[1,:]/aux[0,:]
    return velocity

def stress(current_data):
    """Compute stress from strain and momentum"""
    from clawpack.riemann.nonlinear_elasticity_1D_py import sigma 
    aux=setaux(current_data.x)
    epsilon = current_data.q[0,:]
    stress = sigma(current_data.q,aux,{'stress_relation':'exponential'})
    return stress

if __name__=="__main__":
    from clawpack.pyclaw.util import run_app_from_main
    output = run_app_from_main(setup,setplot)



#!/usr/bin/env python
# encoding: utf-8
"""Output package for PetClaw"""

from __future__ import absolute_import
import logging
from clawpack.pyclaw.fileio import ascii
__all__ = ['ascii.read','ascii.write']

# Check for HDF 5 support
try:
    import h5py
    from . import hdf5
    __all__ += ['hdf5.read','hdf5.write']
except:
    logging.debug("No hdf5 support found.")
    
# Check for netcdf support
try:
    import netCDF4
    from clawpack.pyclaw.fileio import netcdf 
    __all__ += ['netcdf.read','netcdf.write']
except(ImportError):
    logging.debug("No netcdf4 support found.")

# Check for petsc4py support
try:
    from . import petsc
    __all__ += ['petsc.read','petsc.write']
except(ImportError):
    logging.debug("No petsc support found.")
 



#Reconstruction functions for SharpClaw

def weno(k, q):
    import numpy as np

    if k != 5:
        raise ValueError('%d order WENO reconstruction not supported' % k)

    epweno=1.e-36

    dqiph = np.diff(q,1)

    LL=3
    UL=q.shape[1]-2
    qr=q.copy()
    ql=q.copy()

    for m1 in [1,2]:
        #m1=1: construct q^-_{i+1/2} (ql)
        #m1=2: construct q^+_{i+1/2} (qr)
        im=(-1)**(m1+1)
        ione=im
        inone=-im
        intwo=-2*im

        #Create references to DQ slices
        dq_intwo=dqiph[:,LL+intwo-1:UL+intwo-1]
        dq_ione =dqiph[:,LL+ione-1 :UL+ione-1 ]
        dq_inone=dqiph[:,LL+inone-1:UL+inone-1]
        dq      =dqiph[:,LL-1:UL-1            ]

        t1 = im*(dq_intwo-dq_inone)
        t2 = im*(dq_inone-dq)
        t3 = im*(dq      -dq_ione)

        tt1=13.*t1**2+3.*(   dq_intwo - 3.*dq_inone)**2
        tt2=13.*t2**2+3.*(   dq_inone +    dq      )**2
        tt3=13.*t3**2+3.*(3.*dq       -    dq_ione )**2

        tt1=(epweno+tt1)**2
        tt2=(epweno+tt2)**2
        tt3=(epweno+tt3)**2
        s1 = tt2*tt3
        s2 = 6.*tt1*tt3
        s3 = 3.*tt1*tt2
        t0 = 1./(s1+s2+s3)
        s1 *= t0
        s3 *= t0

        z=(s1*(t2-t1)+(0.5*s3-0.25)*(t3-t2))/3. \
                + (-q[:,LL-2:UL-2]+7.*(q[:,LL-1:UL-1]+q[:,LL:UL])-q[:,LL+1:UL+1])/12.
        if m1==1: qr[:,LL-1:UL-1] = z
        else: ql[:,LL:UL] = z

    return ql,qr

def weno5_wave(q,wave,s):

    import numpy as np

    epweno=1.e-36

    qr=q.copy()
    ql=q.copy()
    LL=2
    UL=q.shape[1]-3
    num_waves=wave.shape[1]
    num_eqn=wave.shape[0]
    for m1 in [1,2]:
        #m1=1: construct q^-_{i+1/2} (ql)
        #m1=2: construct q^+_{i+1/2} (qr)
        im=(-1)**(m1+1)
        ione=im
        inone=-im
        intwo=-2*im

        for mw in range(num_waves):
            wnorm2 = wave[0,mw,LL:UL]**2
            theta1 = wave[0,mw,LL+intwo:UL+intwo]*wave[0,mw,LL:UL]
            theta2 = wave[0,mw,LL+inone:UL+inone]*wave[0,mw,LL:UL]
            theta3 = wave[0,mw,LL+ione :UL+ione ]*wave[0,mw,LL:UL]
            for m in range(1,num_eqn):
                wnorm2 += wave[m,mw,LL:UL]**2
                theta1 += wave[m,mw,LL+intwo:UL+intwo]*wave[m,mw,LL:UL]
                theta2 += wave[m,mw,LL+inone:UL+inone]*wave[m,mw,LL:UL]
                theta3 += wave[m,mw,LL+ione :UL+ione ]*wave[m,mw,LL:UL]

            t1=im*(theta1-theta2)
            t2=im*(theta2-wnorm2)
            t3=im*(wnorm2-theta3)

            tt1=13.*t1**2+3.*(   theta1 - 3.*theta2)**2
            tt2=13.*t2**2+3.*(   theta2 +    wnorm2)**2
            tt3=13.*t3**2+3.*(3.*wnorm2 -    theta3)**2

            tt1=(epweno+tt1)**2
            tt2=(epweno+tt2)**2
            tt3=(epweno+tt3)**2
            s1 = tt2*tt3
            s2 = 6.*tt1*tt3
            s3 = 3.*tt1*tt2
            t0 = 1./(s1+s2+s3)
            s1 *= t0
            s3 *= t0

            z=(s1*(t2-t1)+(0.5*s3-0.25)*(t3-t2))/3. \
                    + im*(theta2+6.*wnorm2-theta3)/12.
            u=np.where(wnorm2>1.e-14,z,0.)
            wnorm2=np.where(wnorm2>1.e-14,1./wnorm2,1.)

            for m in range(num_eqn):
                if m1==1: qr[m,LL:UL] += u*wave[m,mw,LL:UL]*wnorm2
                else: ql[m,LL+1:UL+1] += u*wave[m,mw,LL:UL]*wnorm2

    return ql,qr



#!/usr/bin/env python
# encoding: utf-8
r"""
Two-dimensional advection
=========================

Solve the two-dimensional advection equation

.. math::
    q_t + u q_x + v q_y = 0

Here q is a conserved quantity, and (u,v) is the velocity vector.
"""

import numpy as np
from clawpack import riemann


def qinit(state):
    """Set initial condition for q.
       Sample scalar equation with data that is piecewise constant with
       q = 1.0  if  0.1 < x < 0.6   and   0.1 < y < 0.6
           0.1  otherwise
    """
    X, Y = state.grid.p_centers
    state.q[0,:,:] = 0.9*(0.1<X)*(X<0.6)*(0.1<Y)*(Y<0.6) + 0.1
                

def setup(use_petsc=False,outdir='./_output',solver_type='classic',
          dimensional_split=1,transverse_waves=0):

    if use_petsc:
        import clawpack.petclaw as pyclaw
    else:
        from clawpack import pyclaw

    if solver_type == 'classic':
        solver = pyclaw.ClawSolver2D(riemann.advection_2D)
        solver.dimensional_split = dimensional_split
        solver.transverse_waves = transverse_waves
        solver.limiters = pyclaw.limiters.tvd.vanleer
    elif solver_type == 'sharpclaw':
        solver = pyclaw.SharpClawSolver2D(riemann.advection_2D)

    solver.bc_lower[0] = pyclaw.BC.periodic
    solver.bc_upper[0] = pyclaw.BC.periodic
    solver.bc_lower[1] = pyclaw.BC.periodic
    solver.bc_upper[1] = pyclaw.BC.periodic

    if not dimensional_split and not transverse_waves and solver_type == 'classic':
        solver.cfl_max = 0.5
        solver.cfl_desired = 0.45
    else:
        solver.cfl_max = 1.0
        solver.cfl_desired = 0.9

    # Domain:
    mx = 50; my = 50
    x = pyclaw.Dimension(0.0,1.0,mx,name='x')
    y = pyclaw.Dimension(0.0,1.0,my,name='y')
    domain = pyclaw.Domain([x,y])

    num_eqn = 1
    state = pyclaw.State(domain,num_eqn)

    state.problem_data['u'] = 0.5 # Advection velocity
    state.problem_data['v'] = 1.0

    qinit(state)

    claw = pyclaw.Controller()
    claw.tfinal = 2.0
    claw.solution = pyclaw.Solution(state,domain)
    claw.solver = solver
    claw.outdir = outdir
    if outdir is None:
        claw.output_format = None
    claw.setplot = setplot
    claw.keep_copy = True

    return claw


def setplot(plotdata):
    """ 
    Plot solution using VisClaw.
    """ 
    from clawpack.visclaw import colormaps

    plotdata.clearfigures()  # clear any old figures,axes,items data

    # Figure for pcolor plot
    plotfigure = plotdata.new_plotfigure(name='q[0]', figno=0)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'q[0]'
    plotaxes.scaled = True

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = 0
    plotitem.pcolor_cmap = colormaps.yellow_red_blue
    plotitem.pcolor_cmin = 0.0
    plotitem.pcolor_cmax = 1.0
    plotitem.add_colorbar = True
    
    # Figure for contour plot
    plotfigure = plotdata.new_plotfigure(name='contour', figno=1)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'q[0]'
    plotaxes.scaled = True

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_contour')
    plotitem.plot_var = 0
    plotitem.contour_nlevels = 20
    plotitem.contour_min = 0.01
    plotitem.contour_max = 0.99
    plotitem.amr_contour_colors = ['b','k','r']
    
    return plotdata

    
if __name__=="__main__":
    from clawpack.pyclaw.util import run_app_from_main
    output = run_app_from_main(setup,setplot)




""" 
Set up the plot figures, axes, and items to be done for each frame.

This module is imported by the plotting routines and then the
function setplot is called to set the plot parameters.
    
"""

#--------------------------
def setplot(plotdata):
#--------------------------
    
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of clawpack.visclaw.data.ClawPlotData.
    Output: a modified version of plotdata.
    
    """ 

    from clawpack.visclaw import colormaps

    plotdata.clearfigures()  # clear any old figures,axes,items data
    plotdata.format = 'ascii'      # 'ascii', 'binary', 'netcdf'
    

    # Figure for pressure
    # -------------------

    plotfigure = plotdata.new_plotfigure(name='Pressure', figno=0)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = [-4,8]
    plotaxes.ylimits = [-1,11]
    plotaxes.title = 'Pressure'
    plotaxes.scaled = True      # so aspect ratio is 1
    plotaxes.afteraxes = fixup

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = 0
    plotitem.pcolor_cmap = colormaps.blue_white_red
    plotitem.add_colorbar = False
    plotitem.show = True       # show on plot?
    plotitem.pcolor_cmin = -1.5
    plotitem.pcolor_cmax = 1.5
    plotitem.amr_patchedges_show = [1,1,1]
    plotitem.amr_celledges_show = [0,0,0]
    
    # Figure for inner product
    # -------------------
    
    plotfigure = plotdata.new_plotfigure(name='Inner Product', figno=1)
    
    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'Inner Product'
    plotaxes.xlimits = [-4,8]
    plotaxes.ylimits = [-1,11]
    plotaxes.title = 'Inner Product'
    plotaxes.scaled = True      # so aspect ratio is 1
    plotaxes.afteraxes = fixup_innerprod
    
    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = plot_innerprod
    plotitem.pcolor_cmap = colormaps.white_red
    plotitem.add_colorbar = True
    plotitem.show = True       # show on plot?
    #plotitem.pcolor_cmin = 0.01     # use when plotting inner product with q
    plotitem.pcolor_cmin = 0.0      # use when plotting inner product with error
    plotitem.pcolor_cmax = 0.12    # use when plotting inner product with q
    #plotitem.pcolor_cmax = 0.0005    # use when plotting inner product with error
    plotitem.amr_patchedges_show = [0,0,0]
    plotitem.amr_celledges_show = [0,0,0]
    plotitem.amr_data_show = [1,0,0]

    #-----------------------------------------
    # Figures for gauges
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='q', figno=300, \
                    type='each_gauge')
    plotfigure.clf_each_gauge = True

    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = 'auto'
    plotaxes.xlimits = [0,7]
    plotaxes.ylimits = [-0.4,0.5]
    plotaxes.title = 'Pressure'
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = 0
    plotitem.plotstyle = 'b-'
    plotitem.kwargs = {'linewidth': 3}
    plotaxes.afteraxes = fixup_gauge

    # Parameters used only when creating html and/or latex hardcopy
    # e.g., via clawpack.visclaw.frametools.printframes:

    plotdata.printfigs = True                # print figures
    plotdata.print_format = 'png'            # file format
    plotdata.print_framenos = 'all'          # list of frames to print
    plotdata.print_fignos = 'all'            # list of figures to print
    plotdata.html = True                     # create html files of plots?
    plotdata.html_homelink = '../README.html'   # pointer for top of index
    plotdata.html_movie = 'JSAnimation'      # new style, or "4.x" for old style
    plotdata.latex = True                    # create latex file of plots?
    plotdata.latex_figsperline = 2           # layout of plots
    plotdata.latex_framesperline = 1         # layout of plots
    plotdata.latex_makepdf = False           # also run pdflatex?

    return plotdata

def plot_innerprod(current_data):
    return current_data.aux[0,:,:]

# Afteraxis functions:

def addgauges(current_data):
    from clawpack.visclaw import gaugetools
    gaugetools.plot_gauge_locations(current_data.plotdata, \
                                    gaugenos='all', format_string='ko', add_labels=True)

def fixup(current_data):
    import pylab
    addgauges(current_data)

def fixup_innerprod(current_data):
    import pylab
    addgauges(current_data)

def fixup_gauge(current_data):
    import pylab




""" 
Set up the plot figures, axes, and items to be done for each frame.

This module is imported by the plotting routines and then the
function setplot is called to set the plot parameters.
    
""" 

import numpy as np
import matplotlib.pyplot as plt

from clawpack.geoclaw import topotools
import os


#--------------------------
def setplot(plotdata=None):
#--------------------------
    
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of pyclaw.plotters.data.ClawPlotData.
    Output: a modified version of plotdata.
    
    """ 

    import os

    import numpy as np
    import matplotlib.pyplot as plt

    from clawpack.visclaw import geoplot, gaugetools, colormaps

    import clawpack.clawutil.data as clawutil
    import clawpack.amrclaw.data as amrclaw
    import clawpack.geoclaw.data

    import clawpack.geoclaw.multilayer.plot as ml_plot

    if plotdata is None:
        from clawpack.visclaw.data import ClawPlotData
        plotdata = ClawPlotData()

    from numpy import linspace



    plotdata.clearfigures()  # clear any old figures,axes,items data
    plotdata.save_frames = False

    # Load data from output
    clawdata = clawutil.ClawInputData(2)
    clawdata.read(os.path.join(plotdata.outdir,'claw.data'))
    amrdata = amrclaw.AmrclawInputData(clawdata)
    amrdata.read(os.path.join(plotdata.outdir,'amr.data'))
    geodata = clawpack.geoclaw.data.GeoClawData()
    geodata.read(os.path.join(plotdata.outdir,'geoclaw.data'))
    multilayer_data = clawpack.geoclaw.data.MultilayerData()
    multilayer_data.read(os.path.join(plotdata.outdir,'multilayer.data'))

    # To plot gauge locations on pcolor or contour plot, use this as
    # an afteraxis function:

    def addgauges(current_data):
        from clawpack.visclaw import gaugetools
        gaugetools.plot_gauge_locations(current_data.plotdata, \
             gaugenos='all', format_string='ko', add_labels=True)

    # ========================================================================
    #  Generic helper functions
    # ========================================================================
    def pcolor_afteraxes(current_data):
        # bathy_ref_lines(current_data)
        gauge_locations(current_data)
        
    def contour_afteraxes(current_data):
        # gauge_locations(current_data)
        # m_to_km_labels()
        plt.hold(True)
        pos = -80.0 * (23e3 / 180) + 500e3 - 5e3
        plt.plot([pos,pos],[-300e3,300e3],'b',[pos-5e3,pos-5e3],[-300e3,300e3],'y')
        plt.hold(False)
        wind_contours(current_data)
        bathy_ref_lines(current_data)
        
    def profile_afteraxes(current_data):
        pass
        
    def gauge_locations(current_data,gaugenos='all'):
        plt.hold(True)
        gaugetools.plot_gauge_locations(current_data.plotdata, \
             gaugenos=gaugenos, format_string='kx', add_labels=True)
        plt.hold(False)
    
    # ========================================================================
    # Axis limits
    # xlimits = [amrdata.xlower,amrdata.xupper]
    xlimits = [-100.0, 100.0]

    # ylimits = [amrdata.ylower,amrdata.yupper]
    ylimits = [-100.0, 100.0]
    
    eta = [multilayer_data.eta[0],multilayer_data.eta[1]]

    top_surface_limits = [eta[0]-10,eta[0]+10]
    internal_surface_limits = [eta[1]-5,eta[1]+5]
    depth_limits = [0.0, 0.4]
    top_speed_limits = [0.0,0.1]
    internal_speed_limits = [0.0,0.03]
    

    # ========================================================================
    #  Surface Elevations
    # ========================================================================
    plotfigure = plotdata.new_plotfigure(name='Surface', figno=0)
    plotfigure.show = True
    plotfigure.kwargs = {'figsize':(14,4)}
    
    # Top surface
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'Top Surface'
    plotaxes.axescmd = 'subplot(1,2,1)'
    plotaxes.scaled = True
    plotaxes.xlimits = xlimits
    plotaxes.ylimits = ylimits
    ml_plot.add_inundation(plotaxes, 1, bounds=depth_limits)
    ml_plot.add_surface_elevation(plotaxes,1,bounds=top_surface_limits)
    ml_plot.add_land(plotaxes, 1)
    
    # Bottom surface
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'Internal Surface'
    plotaxes.axescmd = 'subplot(1,2,2)'
    plotaxes.scaled = True
    plotaxes.xlimits = xlimits
    plotaxes.ylimits = ylimits
    ml_plot.add_inundation(plotaxes, 2, bounds=depth_limits)
    ml_plot.add_surface_elevation(plotaxes,2,bounds=internal_surface_limits)
    ml_plot.add_colorbar = True
    ml_plot.add_land(plotaxes, 2)


    # ========================================================================
    # Figure for cross section
    # ========================================================================
    plotfigure = plotdata.new_plotfigure(name='cross-section', figno=4)
    plotfigure.show = True
    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = xlimits
    plotaxes.ylimits = ylimits
    plotaxes.title = 'Cross section at y=0'
    ml_plot.add_cross_section(plotaxes, 1)
    ml_plot.add_cross_section(plotaxes, 2)
    ml_plot.add_land_cross_section(plotaxes)


    # ========================================================================
    #  Water Speed
    # ========================================================================
    plotfigure = plotdata.new_plotfigure(name='speed', figno=1)
    plotfigure.show = False
    plotfigure.kwargs = {'figsize':(14,4)}

    # Top layer speed
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'Currents - Top Layer'
    plotaxes.scaled = True
    plotaxes.xlimits = xlimits
    plotaxes.ylimits = ylimits
    plotaxes.axescmd = 'subplot(1,2,1)'
    ml_plot.add_speed(plotaxes,1,bounds=top_speed_limits)
    ml_plot.add_land(plotaxes, 1)
    
    # Bottom layer speed
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = 'Currents - Bottom Layer'
    plotaxes.scaled = True
    plotaxes.xlimits = xlimits
    plotaxes.ylimits = ylimits
    plotaxes.axescmd = 'subplot(1,2,2)'
    ml_plot.add_speed(plotaxes,2,bounds=internal_speed_limits)
    ml_plot.add_land(plotaxes, 2)
    
    # Individual components
    plotfigure = plotdata.new_plotfigure(name='speed_components',figno=401)
    plotfigure.show = False
    plotfigure.kwargs = {'figsize':(14,14)}
    
    # Top layer
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = "X-Velocity - Top Layer"
    plotaxes.scaled = True
    plotaxes.xlimits = xlimits
    plotaxes.ylimits = ylimits
    plotaxes.axescmd = 'subplot(2,2,1)'
    ml_plot.add_x_velocity(plotaxes,1)
    ml_plot.add_land(plotaxes, 1)
    
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = "Y-Velocity - Top Layer"
    plotaxes.scaled = True
    plotaxes.xlimits = xlimits
    plotaxes.ylimits = ylimits
    plotaxes.axescmd = 'subplot(2,2,2)'
    ml_plot.add_y_velocity(plotaxes,1)
    ml_plot.add_land(plotaxes, 1)
    
    # Bottom layer
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = "X-Velocity - Bottom Layer"
    plotaxes.scaled = True
    plotaxes.xlimits = xlimits
    plotaxes.ylimits = ylimits
    plotaxes.axescmd = 'subplot(2,2,3)'
    ml_plot.add_x_velocity(plotaxes,2)
    ml_plot.add_land(plotaxes, 2)
    
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.title = "Y-Velocity - Bottom Layer"
    plotaxes.scaled = True
    plotaxes.xlimits = xlimits
    plotaxes.ylimits = ylimits
    plotaxes.axescmd = 'subplot(2,2,4)'
    ml_plot.add_y_velocity(plotaxes,2)
    ml_plot.add_land(plotaxes, 2)


    #-----------------------------------------
    # Figures for gauges
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='Surface at gauges', figno=300, \
                    type='each_gauge')
    plotfigure.clf_each_gauge = True

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = 'auto'
    plotaxes.title = 'Surface'

    # Plot surface as blue curve:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = 3
    plotitem.plotstyle = 'b-'

    # Plot topo as green curve:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.show = False

    def gaugetopo(current_data):
        q = current_data.q
        h = q[0,:]
        eta = q[3,:]
        topo = eta - h
        return topo
        
    plotitem.plot_var = gaugetopo
    plotitem.plotstyle = 'g-'

    def add_zeroline(current_data):
        from pylab import plot, legend, xticks, floor, axis, xlabel
        t = current_data.t 
        gaugeno = current_data.gaugeno

        if gaugeno == 32412:
            try:
                plot(TG32412[:,0], TG32412[:,1], 'r')
                legend(['GeoClaw','Obs'],loc='lower right')
            except: pass
            axis((0,t.max(),-0.3,0.3))

        plot(t, 0*t, 'k')
        n = int(floor(t.max()/3600.) + 2)
        xticks([3600*i for i in range(n)], ['%i' % i for i in range(n)])
        xlabel('time (hours)')


    #-----------------------------------------
    
    # Parameters used only when creating html and/or latex hardcopy
    # e.g., via pyclaw.plotters.frametools.printframes:

    plotdata.printfigs = True                # print figures
    plotdata.print_format = 'png'            # file format
    plotdata.print_framenos = 'all'          # list of frames to print
    plotdata.print_gaugenos = 'all'          # list of gauges to print
    plotdata.print_fignos = 'all'            # list of figures to print
    plotdata.html = True                     # create html files of plots?
    plotdata.html_homelink = '../README.html'   # pointer for top of index
    plotdata.latex = True                    # create latex file of plots?
    plotdata.latex_figsperline = 2           # layout of plots
    plotdata.latex_framesperline = 1         # layout of plots
    plotdata.latex_makepdf = False           # also run pdflatex?
    plotdata.parallel = True                 # make multiple frame png's at once

    return plotdata




"""
Performs 'make all' in each subdirectory to create sample results for the
gallery or to perform regression tests against results in the gallery,
or elsewhere.

Sends output and errors to separate files to simplify looking for errors.
"""

import os

# Determine directory:
try:
    CLAW = os.environ['CLAW']
except:
    raise Exception("Need to set CLAW environment variable")


def list_examples(examples_dir):
    """
    Searches all subdirectories of examples_dir for examples and prints out a list.
    """
    import os

    current_dir = os.getcwd()
    os.chdir(examples_dir)
    
    dirlist = []
    applist = []

    # Traverse directories depth-first (topdown=False) to insure e.g. that code in
    #    amrclaw/examples/acoustics_2d_radial/1drad 
    # is run before code in
    #    amrclaw/examples/acoustics_2d_radial

    for (dirpath, subdirs, files) in os.walk('.',topdown=False):

        if '_output' not in dirpath:
            # By convention we assume that a setrun.py file indicates this is an
            # example directory.
            files = os.listdir(os.path.abspath(dirpath))
            if 'setrun.py' in files:
                dirlist.append(os.path.abspath(dirpath))

    os.chdir(current_dir)

    return dirlist
        

def make_all(examples_dir = '.',make_clean_first=False, env=None):
    import os,sys

    if env is None:
        my_env = os.environ
    else:
        my_env = env

    examples_dir = os.path.abspath(examples_dir)
    if not os.path.isdir(examples_dir):
        raise Exception("Directory not found: %s" % examples_dir)

    current_dir = os.getcwd()

    dir_list = list_examples(examples_dir)
    print("Found the following example subdirectories:")
    for d in dir_list:
        print("    ", d)
 
    print("Will run code and make plots in the above subdirectories of ")
    print("    ", examples_dir)
    ans = input("Ok? ")
    if ans.lower() not in ['y','yes']:
        print("Aborting.")
        sys.exit()
    
    fname_output = 'make_all_output.txt'
    fout = open(fname_output, 'w')
    fout.write("ALL OUTPUT FROM RUNNING EXAMPLES\n\n")

    fname_errors = 'make_all_errors.txt'
    ferr = open(fname_errors, 'w')
    ferr.write("ALL ERRORS FROM RUNNING EXAMPLES\n\n")

    os.chdir(examples_dir)

    goodlist_run = []
    badlist_run = []
    
    import subprocess
    for directory in dir_list:

        fout.write("\n=============================================\n")
        fout.write(directory)
        fout.write("\n=============================================\n")
        ferr.write("\n=============================================\n")
        ferr.write(directory)
        ferr.write("\n=============================================\n")

        os.chdir(directory)
        print(directory)

        # flush I/O buffers:
        fout.flush()
        ferr.flush()

        if make_clean_first:
            # Run 'make clean':
            job = subprocess.Popen(['make','clean'], \
                      stdout=fout,stderr=ferr)
            return_code = job.wait()
                
        # Run 'make all':
        job = subprocess.Popen(['make','all'], \
                  stdout=fout,stderr=ferr,env=my_env)
        return_code = job.wait()
                
        if return_code == 0:
            print("Successful run\n")
            goodlist_run.append(directory)
        else:
            print("*** Run errors encountered: see %s\n" % fname_errors)
            badlist_run.append(directory)


    print('------------------------------------------------------------- ')
    print(' ')
    print('Ran "make all" and created output and plots in directories:')
    if len(goodlist_run) == 0:
        print('   None')
    else:
        for d in goodlist_run:
            print('   ',d)
    print(' ')
    
    print('Errors encountered in the following directories:')
    if len(badlist_run) == 0:
        print('   None')
    else:
        for d in badlist_run:
            print('   ',d)
    print(' ')
    
    fout.close()
    ferr.close()
    print('For all output see ', fname_output)
    print('For all errors see ', fname_errors)

    os.chdir(current_dir)


def make_notebook_htmls(examples_dir = '.',make_clean_first=False, env=None):
    import os,sys,glob

    if env is None:
        my_env = os.environ
    else:
        my_env = env

    examples_dir = os.path.abspath(examples_dir)
    if not os.path.isdir(examples_dir):
        raise Exception("Directory not found: %s" % examples_dir)

    current_dir = os.getcwd()

    dir_list = list_examples(examples_dir)
    print("Found the following Jupyter notebooks:")
    nb_dir_list = []
    for d in dir_list:
        notebooks = glob.glob(d + '/*.ipynb')
        if len(notebooks) != 0:
            nb_dir_list.append(d)
        for nbfile in notebooks:
            print(nbfile)
            
    if len(nb_dir_list) == 0:
        print("  none")
        sys.exit()
 
    print("Will run notebooks and make htmls in these subdirectories:")
    for nbdir in nb_dir_list:
        print("   ",nbdir)

    ans = input("Ok? ")
    if ans.lower() not in ['y','yes']:
        print("Aborting.")
        sys.exit()
    
    fname_output = 'make_nb_output.txt'
    fout = open(fname_output, 'w')
    fout.write("ALL OUTPUT FROM RUNNING NOTEBOOKS\n\n")

    fname_errors = 'make_nb_errors.txt'
    ferr = open(fname_errors, 'w')
    ferr.write("ALL ERRORS FROM RUNNING NOTEBOOKS\n\n")

    os.chdir(examples_dir)

    goodlist_run = []
    badlist_run = []
    
    import subprocess
    for directory in nb_dir_list:

        fout.write("\n=============================================\n")
        fout.write(directory)
        fout.write("\n=============================================\n")
        ferr.write("\n=============================================\n")
        ferr.write(directory)
        ferr.write("\n=============================================\n")

        os.chdir(directory)

        # flush I/O buffers:
        fout.flush()
        ferr.flush()

        if make_clean_first:
            # Run 'make clean':
            job = subprocess.Popen(['make','clean'], \
                      stdout=fout,stderr=ferr)
            return_code = job.wait()
                
        # Run 'make notebook_htmls':
        job = subprocess.Popen(['make','notebook_htmls'], \
                  stdout=fout,stderr=ferr,env=my_env)
        return_code = job.wait()
                
        if return_code == 0:
            print("Successful run\n")
            goodlist_run.append(directory)
        else:
            print("*** Run errors encountered: see %s\n" % fname_errors)
            badlist_run.append(directory)

        job = subprocess.Popen(['make','README.html'], \
                  stdout=fout,stderr=ferr,env=my_env)
        return_code = job.wait()
        
        if return_code != 0:
            print("*** problems making README.html in %s" % directory)

    print('------------------------------------------------------------- ')
    print(' ')
    print('Ran "make notebook_htmls" in directories:')
    if len(goodlist_run) == 0:
        print('   None')
    else:
        for d in goodlist_run:
            print('   ',d)
    print(' ')
    
    print('Errors encountered in the following directories:')
    if len(badlist_run) == 0:
        print('   None')
    else:
        for d in badlist_run:
            print('   ',d)
    print(' ')
    
    fout.close()
    ferr.close()
    print('For all output see ', fname_output)
    print('For all errors see ', fname_errors)

    os.chdir(current_dir)

if __name__=='__main__':
    import sys
    make_all(*sys.argv[1:])



r"""
xarray backends module: $CLAW/geoclaw/src/python/geoclaw/xarray_backends.py

Xarray backends for GeoClaw fixed grids and fgmax grids.

These only work for point_style = 2 (uniform regular) and have a dependency on
xarray and rioxarray. Xarray provides the core datastructure and rioxarray
provides an interface to rasterio, used to assign geospatial projection
information.

- https://docs.xarray.dev/en/stable/index.html
- https://corteva.github.io/rioxarray/stable/readme.html
- https://rasterio.readthedocs.io/en/latest/index.html

The expectation is that you run commands that use these backends from the same
directory you run "make output".

Includes:

- class FGMaxBackend: Xarray backend for fgmax grids.
- class FGOutBackend: Xarray backend for fgout grids.

Usage:

.. code-block:: python
    import rioxarray  # activate the rio accessor
    import xarray as xr
    from clawpack.geoclaw.xarray_backends import FGOutBackend, FGMaxBackend

    # An example of a fgout grid.

    filename = '_output/fgout0001.b0001
    # provide the .bxxx file if binary format is used or the
    # .qxxx file if ascii format is used.
    # the format, fg number, and frame number are inferred from the filename.

    ds = xr.open_dataset(
        filename,
        engine=FGOutBackend,
        backend_kwargs={
            'qmap': 'geoclaw',
            'epsg':epsg_code
                })
    # ds is now an xarray object. It can be interacted with directly or written to netcdf using
    ds.write_netcdf('filename.nc')

    # A 'qmap' backend_kwargs is required as it indicates the qmap used for
    # selecting elements of q to include in fgout. See
    # https://www.clawpack.org/dev/fgout.html#specifying-q-out-vars

    # Optionally, provide an epsg code to assign the associated coordinate system to the file.
    # default behavior assigns no coordinate system.

    # An example of a fgmax grid.
    filename = "_output/fgmax0001.txt"
    ds = xr.open_dataset(filename, engine=FGMaxBackend, backend_kwargs={'epsg':epsg_code})


Dimensions:

Files opened with FGOutBackend will have dimensions (time, y, x).
Files opened with FGMaxBackend will have dimensions (y, x).

Variable naming:

For fixed grid files, the dataset will have the elements of q specified
by qmap and q_out_vars. This may be as many as the full list described below or
as few as specified by the user.

For fixed grid geoclaw files, the full set of variables is:
- h
- hu
- hv
- eta
- B

For fixed grid geoclaw-bouss files, the full set of variables is:
- h
- hu
- hv
- huc
- hvc
- eta
- B

Fixed grid dclaw files, the full set of variables is:
- h
- hu
- hv
- hm
- pb
- hchi
- bdif
- eta
- B

Depending on the number of variables specified in the setrun.py fgmax files will
have a portion of the following variables:

If rundata.fgmax_data.num_fgmax_val == 1

- arrival_time, Wave arrival time (based on eta>sea_level + fg.arrival_tol)
- h_max, Maximum water depth
- eta_max, Maximum water surface elevation
- h_max_time, Time of maximum water depth
- B, Basal topography at the first time fgmax first monitored maximum amr level
- level, Maximum amr level

If rundata.fgmax_data.num_fgmax_val == 2:

- s_max, Maximum velocity
- s_max_time, Time of maximum velocity

If rundata.fgmax_data.num_fgmax_val == 5:

- hs_max, Maximum momentum
- hs_max_time, Time of maximum momentum
- hss_max, Maximum momentum flux
- hss_max_time, Time of maximum momentum flux
- h_min, Minimum depth
- h_min_time, Time of minimum depth


See the following links for additional information about xarray Backends.

- https://docs.xarray.dev/en/stable/generated/xarray.backends.BackendEntrypoint.html#xarray.backends.BackendEntrypoint
- https://docs.xarray.dev/en/stable/generated/xarray.open_dataset.html
"""

import os

import numpy as np

try:
    import rioxarray  # activate the rio accessor
    import xarray as xr
except ImportError:
    raise ImportError(
        "rioxarray and xarray are required to use the FGOutBackend and FGMaxBackend"
    )

from clawpack.geoclaw import fgmax_tools, fgout_tools
from xarray.backends import BackendEntrypoint

_qunits = {
    "h": "meters",
    "hu": "meters squared per second",
    "hv": "meters squared per second",
    "hm": "meters",
    "pb": "newton per meter squared",
    "hchi": "meters",
    "bdif": "meters",
    "eta": "meters",
    "B": "meters",
    "huc": "varies",
    "hvc": "varies",
}

time_unit = "seconds"
reference_time = "model start"
space_unit = "meters"
nodata = np.nan


def _prepare_var(data, mask, dims, units, nodata, long_name):
    "Reorient array into the format xarray expects and generate the mapping."
    if mask is not None:
        data[mask] = nodata
    data = data.T
    data = np.flipud(data)
    mapping = (
        dims,
        data,
        {"units": units, "_FillValue": nodata, "long_name": long_name},
    )
    return mapping


class FGOutBackend(BackendEntrypoint):
    "Xarray Backend for Clawpack fixed grid format."

    def open_dataset(
        self,
        filename,  # path to fgout file.
        qmap="geoclaw",  # qmap value for FGoutGrid ('geoclaw', 'dclaw', or 'geoclaw-bouss')
        epsg=None,  # epsg code
        dry_tolerance=0.001,
        # dry tolerance used for for masking elements of q that are not
        # eta or B. Default behavior is to mask all elements of q except for
        # eta and B where h>0.001.
        # used only if h or eta-B is available based on q_out_vars.
        # if dry_tolerance = None, no masking is applied to any variable.
        drop_variables=None,  # name of any elements of q to drop.
    ):

        if drop_variables is None:
            drop_variables = []

        full_path = os.path.abspath(filename)
        filename = os.path.basename(full_path)
        outdir = os.path.basename(os.path.dirname(full_path))

        # filename has the format fgoutXXXX.qYYYY (ascii)
        # or fgoutXXXX.bYYYY (binary)
        # where XXXX is the fixed grid number and YYYY is the frame
        # number.
        type_code = filename.split(".")[-1][0]
        fgno = int(filename.split(".")[0][-4:])
        frameno = int(filename.split(".")[-1][-4:])
        if type_code == "q":  # TODO, is this correct?
            output_format = "ascii"
        elif type_code == "b":
            output_format = "binary32"  # format of fgout grid output
        else:
            raise ValueError("Invalid FGout output format. Must be ascii or binary.")

        fgout_grid = fgout_tools.FGoutGrid(
            fgno=fgno, outdir=outdir, output_format=output_format, qmap=qmap
        )

        if fgout_grid.point_style != 2:
            raise ValueError("FGOutBackend only works with fg.point_style=2")

        fgout_grid.read_fgout_grids_data()
        fgout = fgout_grid.read_frame(frameno)

        time = fgout.t
        # both come in ascending. flip to give expected order.
        x = fgout.x
        y = np.flipud(fgout.y)
        nj = len(x)
        ni = len(y)

        # mask based on dry tolerance
        if dry_tolerance is not None:
            try:
                mask = fgout.h < dry_tolerance
                # internally fgout_tools will try fgou.eta-fgout.B if h is not present.
            except AttributeError:
                print("FGOutBackend: No h, eta, or B. No mask applied.")
                mask = np.ones((nj, ni), dtype=bool)

        # create data_vars dictionary
        data_vars = {}
        for i, i_var in enumerate(fgout_grid.q_out_vars):

            # construct variable

            # Find the varname in fgout.qmap associated with
            # q_out_vars[i]
            varname = None
            for name, index in fgout_grid.qmap.items():
                if i_var == index:
                    varname = name

            # construct xarray dataset if varname not in drop vars.
            if varname not in drop_variables:

                Q = fgout.q[i, :, :]

                # mask all but eta based on h presence.
                if (varname not in ("B", "eta")) and dry_tolerance is not None:
                    Q[mask] = nodata

                # to keep xarray happy, need to transpose and flip ud.
                Q = Q.T
                Q = np.flipud(Q)
                Q = Q.reshape((1, ni, nj))  # reshape to add a time dimension.

                data_array_attrs = {"units": _qunits[varname], "_FillValue": nodata}

                data_vars[varname] = (
                    [
                        "time",
                        "y",
                        "x",
                    ],
                    Q,
                    data_array_attrs,
                )

        ds_attrs = {"description": "Clawpack model output"}

        ds = xr.Dataset(
            data_vars=data_vars,
            coords=dict(
                x=(["x"], x, {"units": space_unit}),
                y=(["y"], y, {"units": space_unit}),
                time=("time", [time], {"units": "seconds"}),
                reference_time=reference_time,
            ),
            attrs=ds_attrs,
        )

        if epsg is not None:
            ds.rio.write_crs(
                epsg,
                inplace=True,
            ).rio.set_spatial_dims(
                x_dim="x",
                y_dim="y",
                inplace=True,
            ).rio.write_coordinate_system(inplace=True)
            # https://corteva.github.io/rioxarray/stable/getting_started/crs_management.html#Spatial-dimensions
            # https://gis.stackexchange.com/questions/470207/how-to-write-crs-info-to-netcdf-in-a-way-qgis-can-read-python-xarray

        return ds

    open_dataset_parameters = ["filename", "drop_variables"]

    description = "Use Clawpack fixed grid output files in Xarray"
    url = "https://www.clawpack.org/fgout.html"


class FGMaxBackend(BackendEntrypoint):
    "Xarray Backend for Clawpack fgmax grid format."

    def open_dataset(
        self,
        filename,
        epsg=None,
        drop_variables=None,
        clip=False,  # if True, clip the entire array to the extent where arrival_time is defined.
    ):

        if drop_variables is None:
            drop_variables = []

        # expectation is for standard clawpack organization,
        # e.g., output and 'fgmax_grids.data' in _output
        fgno = int(os.path.basename(filename).split(".")[0][-4:])
        outdir = os.path.dirname(filename)
        data_file = os.path.join(outdir, "fgmax_grids.data")

        fg = fgmax_tools.FGmaxGrid()
        fg.read_fgmax_grids_data(fgno=fgno, data_file=data_file)
        if fg.point_style != 2:
            raise ValueError("FGMaxBackend only works with fg.point_style=2")

        fg.read_output(outdir=outdir)

        # Construct the x and y coordinates
        # Both come in ascending, therefore flip y so that it is ordered as expected.
        x = fg.x
        y = np.flipud(fg.y)

        # Construct the data_vars array. To organize the
        # data in the way expected by netcdf standards, need
        # to both transpose and flipud the array.

        data_vars = {}

        data_vars["arrival_time"] = _prepare_var(
            fg.arrival_time.data,
            fg.arrival_time.mask,
            [
                "y",
                "x",
            ],
            "seconds",
            nodata,
            "Wave arrival time",
        )

        data_vars["h_max"] = _prepare_var(
            fg.h.data,
            fg.h.mask,
            [
                "y",
                "x",
            ],
            "meters",
            nodata,
            "Maximum water depth",
        )

        data_vars["eta_max"] = _prepare_var(
            fg.h.data + fg.B.data,
            fg.h.mask,
            [
                "y",
                "x",
            ],
            "meters",
            nodata,
            "Maximum water surface elevation",
        )

        data_vars["h_max_time"] = _prepare_var(
            fg.h_time.data,
            fg.h_time.mask,
            [
                "y",
                "x",
            ],
            "seconds",
            nodata,
            "Time of maximum water depth",
        )

        data_vars["B"] = _prepare_var(
            fg.B,
            None,
            [
                "y",
                "x",
            ],
            "meters",
            nodata,
            "Basal topography at the first time fgmax first monitored maximum amr level",
        )

        data_vars["level"] = _prepare_var(
            fg.level,
            None,
            [
                "y",
                "x",
            ],
            "(no units)",
            -1,
            "Maximum amr level",
        )

        if hasattr(fg, "s"):
            if hasattr(fg.s, "data"):
                data_vars["s_max"] = _prepare_var(
                    fg.s.data,
                    fg.s.mask,
                    [
                        "y",
                        "x",
                    ],
                    "meters per second",
                    nodata,
                    "Maximum velocity",
                )
                data_vars["s_max_time"] = _prepare_var(
                    fg.s_time.data,
                    fg.s_time.mask,
                    [
                        "y",
                        "x",
                    ],
                    "seconds",
                    nodata,
                    "Time of maximum velocity",
                )
        if hasattr(fg, "hs"):
            if hasattr(fg.hs, "data"):
                data_vars["hs_max"] = _prepare_var(
                    fg.hs.data,
                    fg.hs.mask,
                    [
                        "y",
                        "x",
                    ],
                    "meters squared per second",
                    nodata,
                    "Maximum momentum",
                )
                data_vars["hs_max_time"] = _prepare_var(
                    fg.hs_time.data,
                    fg.hs_time.mask,
                    [
                        "y",
                        "x",
                    ],
                    "seconds",
                    nodata,
                    "Time of maximum momentum",
                )

                data_vars["hss_max"] = _prepare_var(
                    fg.hss.data,
                    fg.hss.mask,
                    [
                        "y",
                        "x",
                    ],
                    "meters cubed per second squared",
                    nodata,
                    "Maximum momentum flux",
                )
                data_vars["hss_max_time"] = _prepare_var(
                    fg.hss_time.data,
                    fg.hss_time.mask,
                    [
                        "y",
                        "x",
                    ],
                    "seconds",
                    nodata,
                    "Time of maximum momentum flux",
                )

                data_vars["h_min"] = _prepare_var(
                    fg.hmin.data,
                    fg.hmin.mask,
                    [
                        "y",
                        "x",
                    ],
                    "meters",
                    nodata,
                    "Minimum depth",
                )
                data_vars["h_min_time"] = _prepare_var(
                    fg.hmin_time.data,
                    fg.hmin_time.mask,
                    [
                        "y",
                        "x",
                    ],
                    "seconds",
                    nodata,
                    "Time of minimum depth",
                )

        # drop requested variables
        for var in drop_variables:
            if var in data_vars.keys():
                del data_vars[var]

        # Construct the values from
        ds_attrs = {"description": "D-Claw model output"}

        ds = xr.Dataset(
            data_vars=data_vars,
            coords=dict(
                x=(["x"], x, {"units": "meters"}),
                y=(["y"], y, {"units": "meters"}),
            ),
            attrs=ds_attrs,
        )

        if epsg is not None:

            ds.rio.write_crs(
                epsg,
                inplace=True,
            ).rio.set_spatial_dims(
                x_dim="x",
                y_dim="y",
                inplace=True,
            ).rio.write_coordinate_system(inplace=True)
            # https://corteva.github.io/rioxarray/stable/getting_started/crs_management.html#Spatial-dimensions
            # https://gis.stackexchange.com/questions/470207/how-to-write-crs-info-to-netcdf-in-a-way-qgis-can-read-python-xarray

        # clip
        if clip:
            clip_data = fg.arrival_time.data >= 0
            clip_data = clip_data.T
            clip_data = np.flipud(clip_data)
            ds = _clip(ds, clip_data)

        return ds

    open_dataset_parameters = ["filename", "drop_variables", "clip"]

    description = "Use Clawpack fix grid monitoring files in Xarray"
    url = "https://www.clawpack.org/fgmax.html#fgmax"


def _clip(ds, clip_data):
    # clip DS to the extent where clip_data is true.
    # useful for when there are large areas where nothing happened.

    nrow, ncol = clip_data.shape

    valid_col = np.sum(clip_data, axis=0) > 0
    valid_row = np.sum(clip_data, axis=1) > 0

    sel_rows = np.nonzero(valid_row)[0]
    sel_cols = np.nonzero(valid_col)[0]

    first_row = sel_rows[0]
    last_row = sel_rows[-1]
    first_col = sel_cols[0]
    last_col = sel_cols[-1]

    ds = ds.isel(y=np.arange(first_row, last_row), x=np.arange(first_col, last_col))
    return ds



"""
Plot timing info found in timing.csv file.
Requires modified valout function from Clawpack 5.5.0 to print this info.

This might eventually be turned into a more general utility function in visclaw.
For now, copy this file and modify it for your needs.
"""

from __future__ import print_function
from pylab import *
import os

# Location of timing.csv files:
outdir = '_output'

make_pngs = True  # print plots?

def make_png(fname):
    savefig(fname)
    #savefig(fname, bbox_inches='tight')
    print('Created %s' % fname)

# set desired units for simulation time and computer time,
# based on length of run:

# set units and scaling factors for Simulation time t (simtime), 
# CPU/Wall time (comptime), and for number of cells updated.
# units is used for text in plots, factor is used for scaling values read in
# if None, then no units appear in plots, and/or no scaling is done.

# computer time is recorded in seconds so this is the natural units:
comptime_units = 'seconds'
if comptime_units == 'seconds':
    comptime_factor = 1
elif comptime_units == 'minutes':
    comptime_factor = 60.
elif comptime_units == 'hours':
    comptime_factor = 3600.
else:
    comptime_factor = 1

# for applications where t in the PDE is dimensionless, use this:
simtime_units = 'dimensionless'
simtime_factor = 1

# for GeoClaw or other applications where the simulation time is in
# seconds, set simtime_units to 'seconds', 'minutes' or 'hours' depending
# on time scale of simulation and use this:
#simtime_units = 'seconds'
if simtime_units == 'seconds':
    simtime_factor = 1
elif simtime_units == 'minutes':
    simtime_factor = 60.
elif simtime_units == 'hours':
    simtime_factor = 3600.

# if the time units in the application are different, set simtime_units
# and simtime_factor appropriately.
    
# Some useful units for cell updates:
#cell_units = None # for raw cell count
cell_units = 'millions'

if cell_units == 'millions':
    cell_factor = 1e6
elif cell_units == 'billions':
    cell_factor = 1e9
elif cell_units == 'trillions':
    cell_factor = 1e12
else:
    cell_factor = 1


# define colors, with colors[0] used for overhead, colors[j] for level j >= 1
colors = ['y'] + 3*['r','g','m','c','b']  # allow <= 15 levels


# Read in timing states from output directory:
timing_stats_file = os.path.join(outdir, 'timing.csv')
timing_stats = loadtxt(timing_stats_file, skiprows=1, delimiter=',')
ntimes = timing_stats.shape[0]
nlevels = int(timing_stats.shape[1] / 3) - 1

time = zeros(ntimes)
total_cpu = zeros(ntimes)
total_wall = zeros(ntimes)
wtime = zeros((ntimes, nlevels))
cpu = zeros((ntimes, nlevels))
cells = zeros((ntimes, nlevels))

for j in range(ntimes):
    time[j] = timing_stats[j,0] / simtime_factor
    total_wall[j] = timing_stats[j,1] / comptime_factor
    total_cpu[j] = timing_stats[j,2] / comptime_factor
    for level in range(nlevels):
        wtime[j, level] = timing_stats[j,3*level + 3] / comptime_factor
        cpu[j, level] = timing_stats[j,3*level + 4] / comptime_factor
        cells[j, level] = timing_stats[j,3*level + 5] / cell_factor
    
xlimits = [time.min(), time.max()]
ylimits_comptime = [0, 1.1*total_cpu.max()]
ylimits_cells = [0, 1.1*sum(cells[-1,:])]


figure(21)
clf()
sum_cells_over_levels = zeros(ntimes)
for j in range(nlevels):
    if max(cells[:,j]) == 0:
        break
    #plot(time/3600, cells[:,j], label='Level %s' % (j+1))
    last_sum_cells = sum_cells_over_levels.copy()
    sum_cells_over_levels += cells[:,j]
    plot(time, sum_cells_over_levels, 'k')
    fill_between(time, last_sum_cells, sum_cells_over_levels, 
                 color=colors[j+1],
                 label='Level %s' % (j+1))

plot(time, sum_cells_over_levels, 'k', lw=3, label='Total Cells')
xlim(xlimits)
#ylim(0, 1.1*sum_cells_over_levels[-1])
ylim(ylimits_cells)
title('Cumulative cells updated on each level')
xlabel('Simulation time t (%s)' % simtime_units)
if cell_units is None:
    ylabel('Grid cells updated')
else:
    ylabel('Grid cells updated (%s)' % cell_units)

legend(loc='upper left')

if make_pngs:
    make_png('CumCellUpdates.png')


figure(22)
clf()
sum_cpu_over_levels = zeros(ntimes)
for j in range(nlevels):
    if max(cpu[:,j]) == 0:
        break
    #plot(time/3600, cpu[:,j], label='Level %s' % (j+1))
    last_sum_cpu = sum_cpu_over_levels.copy()
    sum_cpu_over_levels += cpu[:,j]
    fill_between(time, last_sum_cpu, sum_cpu_over_levels, color=colors[j+1],
                 label='Level %s' % (j+1))
    plot(time, sum_cpu_over_levels, 'k')

fill_between(time, total_cpu,sum_cpu_over_levels,color=colors[0],
             label='Overhead')
plot(time, total_cpu, 'k', lw=3, label='Total CPU')
xlim(xlimits)
ylim(ylimits_comptime)
title('Cumulative CPU time on each level')
xlabel('Simulation time t (%s)' % simtime_units)
ylabel('CPU time (%s)' % comptime_units)
legend(loc='upper left')

if make_pngs:
    make_png('CumCPUtime.png')


figure(23)
clf()
sum_wtime_over_levels = zeros(ntimes)
for j in range(nlevels):
    if max(wtime[:,j]) == 0:
        break
    last_sum_wtime = sum_wtime_over_levels.copy()
    sum_wtime_over_levels += wtime[:,j]
    fill_between(time, last_sum_wtime,
sum_wtime_over_levels, 
color=colors[j+1],
                 label='Level %s' % (j+1))
    plot(time, sum_wtime_over_levels, 'k')

fill_between(time, total_wall, sum_wtime_over_levels, color=colors[0],
             label='Overhead')
plot(time, total_wall, 'k', lw=3, label='Total Wall')
title('Cumulative wall time on each level')
xlabel('Simulation time t (%s)' % simtime_units)
ylabel('CPU time (%s)' % comptime_units)
legend(loc='upper left')
xlim(xlimits)
ylim(ylimits_comptime)

if make_pngs:
    make_png('CumWallTime.png')


# d cells / dt:

figure(31)
clf()
dc_max = 0
dca = cells[1:,:] - cells[:-1,:]
for n in range(1,ntimes):
    dt = (time[n] - time[n-1])
    if dt == 0:
        break
    dcn = 0
    for j in range(nlevels):
        if dca[n-1,j] == 0:
            break
        tt = array([time[n-1],time[n]])
        #last_dc = last_dc + dc
        dc = (cells[n,j] - cells[n-1,j]) / dt
        plot(tt, [dcn+dc, dcn+dc], 'k')
        if n == 1:
            fill_between(tt, [dcn,dcn], [dcn+dc,dcn+dc],
                         color=colors[j+1],
                         label='Level %s' % (j+1))
        else:
            fill_between(tt, [dcn,dcn], [dcn+dc,dcn+dc],
                         color=colors[j+1])
        dcn = dcn + dc
                         
    plot([time[n-1],time[n-1]], [0,dcn], 'k')
    plot([time[n],time[n]], [0,dcn], 'k')
    dc_max = max(dc_max, dcn)

                         
xlim(xlimits)
ylim(0, 1.2*dc_max)
title('Average Cells updated / simulation time')
xlabel('Simulation time t (%s)' % simtime_units)
ylabel('cell updates / sim time')
if cell_units is None:
    ylabel('Grid cells updated / sim time')
else:
    ylabel('Grid cells updated (%s) / sim time' % cell_units)

legend(loc='upper left')

if make_pngs:
    make_png('AveCellUpdates.png')


# average cpu_time / dt:

figure(32)
clf()
dc_max = 0
dca = cpu[1:,:] - cpu[:-1,:]
for n in range(1,ntimes):
    dt = (time[n] - time[n-1])
    if dt == 0:
        break
    dcn = 0
    for j in range(nlevels):
        if dca[n-1,j] == 0:
            break
        tt = array([time[n-1],time[n]])
        #last_dc = last_dc + dc
        dc = (cpu[n,j] - cpu[n-1,j]) / dt
        plot(tt, [dcn+dc, dcn+dc], 'k')
        if n == 1:
            fill_between(tt, [dcn,dcn], [dcn+dc,dcn+dc],
                         color=colors[j+1],
                         label='Level %s' % (j+1))
        else:
            fill_between(tt, [dcn,dcn], [dcn+dc,dcn+dc],
                         color=colors[j+1])
        dcn = dcn + dc
                         

    
    if n == 1:
        kwargs_label = {'label': 'Overhead'}
    else:
        kwargs_label = {}
    dtot = (total_cpu[n]-total_cpu[n-1]) / dt 
    plot(tt, [dtot,dtot], 'k')
    fill_between(tt, [dcn,dcn], [dtot,dtot], 
                 color=colors[0], alpha=1, edgecolors='k', **kwargs_label)

    plot([time[n-1],time[n-1]], [0,dtot], 'k')
    plot([time[n],time[n]], [0,dtot], 'k')
    
    dc_max = max(dc_max, dtot)
     
    #plot(time, total_cpu, 'k', lw=3, label='Total CPU')

                         
#plot(time, sum_cells_over_levels, 'k', lw=3, label='Total Cells')
xlim(xlimits)
ylimits_avecomptime = [0, 1.2*dc_max]
ylim(ylimits_avecomptime)

title('Average CPU time / simulation time')
xlabel('Simulation time t (%s)' % simtime_units)
ylabel('CPU time / sim time')
legend(loc='upper left')

if make_pngs:
    make_png('AveCPUTime.png')



# average wall time / dt:

figure(33)
clf()
dc_max = 0
dca = wtime[1:,:] - wtime[:-1,:]
for n in range(1,ntimes):
    dt = (time[n] - time[n-1])
    if dt == 0:
        break
    dcn = 0
    for j in range(nlevels):
        if dca[n-1,j] == 0:
            break
        tt = array([time[n-1],time[n]])
        #last_dc = last_dc + dc
        dc = (wtime[n,j] - wtime[n-1,j]) / dt
        plot(tt, [dcn+dc, dcn+dc], 'k')
        if n == 1:
            fill_between(tt, [dcn,dcn], [dcn+dc,dcn+dc],
                         color=colors[j+1],
                         label='Level %s' % (j+1))
        else:
            fill_between(tt, [dcn,dcn], [dcn+dc,dcn+dc],
                         color=colors[j+1])
        dcn = dcn + dc
                         

    
    if n == 1:
        kwargs_label = {'label': 'Overhead'}
    else:
        kwargs_label = {}
    dtot = (total_wall[n]-total_wall[n-1]) / dt 
    plot(tt, [dtot,dtot], 'k')
    fill_between(tt, [dcn,dcn], [dtot,dtot], 
                 color=colors[0], alpha=1, edgecolors='k', **kwargs_label)

    plot([time[n-1],time[n-1]], [0,dtot], 'k')
    plot([time[n],time[n]], [0,dtot], 'k')
    
    dc_max = max(dc_max, dtot)

                         
xlim(xlimits)
#ylim(0, 1.2*dc_max)
ylim(ylimits_avecomptime)
title('Average Wall time / simulation time')
xlabel('Simulation time t (%s)' % simtime_units)
ylabel('wtime time / sim time')
legend(loc='upper left')

if make_pngs:
    make_png('AveWallTime.png')



from . import quadrants, shock_bubble_interaction, shock_forward_step
import numpy as np
from pytest import importorskip
import os

thisdir = os.path.dirname(__file__)
path_expected_sols = os.path.join(thisdir,'expected_sols.npy')
expected_sols_dict = np.load(path_expected_sols,allow_pickle=True).item()

def error(test_name,**kwargs):
    if "quadrants" in test_name:
        claw = quadrants.setup(**kwargs)
    elif "shock_bubble_interaction" in test_name:
        claw = shock_bubble_interaction.setup(**kwargs)
    elif "shock_forward_step" in test_name:
        claw = shock_forward_step.setup(**kwargs)
    else:
        raise ValueError(f"Test name {test_name} not recognized")
    claw.run()
    density_test = claw.frames[claw.num_output_times].state.get_q_global()[0].reshape([-1])
    dx = claw.solution.domain.grid.delta[0]
    dy = claw.solution.domain.grid.delta[1]
    assert test_name in expected_sols_dict.keys(), f"Test name {test_name} not found in {path_expected_sols}"
    density_expected = expected_sols_dict[test_name]
    diff_L1 = dx*dy*np.sum(np.abs(density_test-density_expected))
    return diff_L1

class TestEuler2D:
    def test_quadrants_hlle(self):
        assert error(test_name='quadrants_hlle', riemann_solver='hlle')<1e-6

    def test_quadrants_roe(self):
        assert error(test_name='quadrants_roe', riemann_solver='roe')<1e-6

    def test_shock_bubble_interaction(self):
        _ = importorskip("scipy") # skip test if scipy is not installed
        assert error(test_name='shock_bubble_interaction', mx=160, my=40, tfinal=0.2, num_output_times=1, disable_output=True)<1e-6
    
    # def test_shock_forward_step(self):
    #     assert error(test_name='shock_forward_step',tfinal=0.1)<1e-6


r"""
Data Module

Contains the general class definition and the subclasses of the Clawpack data
objects.

Changes in 5.0:
 Stripped down version of Data object, no longer keeps track of "owners".
 Data classes are now subclasses of ClawData, which checks if attributes
   already exist before setting.

"""

import os
import sys
import shutil
import inspect
import tarfile
import zipfile
import gzip
import bz2
import string

try:
    from urllib.request import urlopen
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen

import numpy as np

# ======================
#  Remote file handling
# ======================
def strip_archive_extensions(path, 
                             extensions=["tar", "tgz", "bz2", "gz", "zip"]):
    r"""
    Strip off archive extensions defined in *extensions* list.

    Return stripped path calling this function recursively until all splitext
    does not provide an extension in the *extensions* list.  Note that zip
    files store the names of the contained files in the archive and the 
    returned path will more than likely not have the appropriate file suffix.

    """

    if os.path.splitext(path)[-1][1:] in extensions:
        return strip_archive_extensions(os.path.splitext(path)[0])
    else:
        return path


def get_remote_file(url, output_dir=None, file_name=None, force=False,  
                         verbose=False, ask_user=False, unpack=True):
    r"""Fetch file located at *url* and store at *output_dir*.

    :Input:
    
     - *url* (path) - URL to file to be downloaded.
     - *output_dir* (path) - Directory that the remote file will be downloaded
       to.  Defaults to the GeoClaw sratch directory defined by 
       *os.path.join(os.environ['CLAW'], 'geoclaw', 'scratch')*.
     - *file_name* (string) - Name of local file.  This defaults to the name of
       the remote file.
     - *force* (bool) - Force downloading of remote file regardless of whether
       it exists locally or not.  Default is *False*
     - *verbose* (bool) - Print out status information.  Default is *False*
     - *ask_user* (bool) - Whether to ask the user if it is ok to download the
       file before proceeding.  Default is *False*

    :Raises:
     
    Exceptions are raised from the *urllib* module having to do with errors
    fetching the remote file.  Please see its documentation for more details of
    the exceptions that can be raised.

    :Notes:

    If fetching the file fails *urllib* does not always warn you so if errors
    occur it may be wise to check to make sure the downloaded file actually has
    content in it.

    returns *unarchived_output_path*
    """

    if output_dir is None:
        output_dir = os.path.join(os.environ['CLAW'], 'geoclaw', 'scratch')

    if file_name is None:
        file_name = os.path.basename(url)
        
    output_path = os.path.join(output_dir, file_name)
    unarchived_output_path = strip_archive_extensions(output_path)

    if not os.path.exists(unarchived_output_path) or force:

        if ask_user:
            ans = input("  Ok to download topo file and save as %s?  \n"
                            % unarchived_output_path,
                            "     Type y[es], n[o].")
            if ans.lower() in ['y', 'yes']:
                if verbose:
                    print("*** Aborting download.")
                return None
            
        # if not os.path.exists(output_path):
        # Fetch remote file, will raise a variety of exceptions depending on
        # the retrieval problem if it happens
        if verbose:
            print("Downloading %s to %s..." % (url, output_path))
        with open(output_path, "wb") as output_file:
            remote_file = urlopen(url)
            shutil.copyfileobj(remote_file, output_file)
        if verbose:
            print("Done downloading.")
        # elif verbose:
        #     print("File already exists, not downloading")

        if tarfile.is_tarfile(output_path) and unpack:
            if verbose:
                print("Un-archiving %s to %s..." % (output_path, 
                                                    unarchived_output_path))
            with tarfile.open(output_path, mode="r:*") as tar_file:
                def is_within_directory(directory, target):
                    
                    abs_directory = os.path.abspath(directory)
                    abs_target = os.path.abspath(target)
                
                    prefix = os.path.commonprefix([abs_directory, abs_target])
                    
                    return prefix == abs_directory
                
                def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
                
                    for member in tar.getmembers():
                        member_path = os.path.join(path, member.name)
                        if not is_within_directory(path, member_path):
                            raise Exception("Attempted Path Traversal in Tar File")
                
                    tar.extractall(path, members, numeric_owner=numeric_owner) 
                    
                
                safe_extract(tar_file, path=output_dir)
            if verbose:
                print("Done un-archiving.")
        elif zipfile.is_zipfile(output_path) and unpack:
            if verbose:
                print("Un-archiving %s to %s..." % (output_path, 
                                                    unarchived_output_path))
            with zipfile.ZipFile(output_path, mode="r") as zip_file:
                zip_file.extractall(path=output_dir)
                # Add file suffix
                extension = os.path.splitext(zip_file.namelist()[0])[-1]
                unarchived_output_path = "".join((unarchived_output_path,
                                                  extension))
            if verbose:
                print("Done un-archiving.")
        elif os.path.splitext(output_path)[-1] in [".gz",".bz2"] and unpack:
            unzipper = gzip if output_path[-2:] == "gz" else bz2
            if verbose:
                print("Un-archiving %s to %s..." % (output_path,
                                                    unarchived_output_path))
            with unzipper.open(output_path, "r") as f_in:
                with open(unarchived_output_path, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            if verbose:
                print("Done un-archiving.")

    else:
        if verbose:
            print("Skipping %s " % url)
            print("  because file already exists: %s" % output_path)

    if unpack:
        return unarchived_output_path
    else:
        return output_path



# ==============================================================================
#  Base data class for Clawpack data objects
class ClawData(object):
    r"""
    Class to be subclassed when defining data objects that should have
    a limited set of allowed attributes.  Useful to guard against
    typos or misrembering the names of expected attributes.

    Resetting values of existing attributes is allowed as usual,
    but new attributes can only be added using the method add_attribute.

    Trying to set a nonexistent attribute will raise an AttributeError
    exception, except for those starting with '_'.   
    """


    def __init__(self, attributes=None):
        
        # Attribute to store a list of the allowed attributes, 
        # appended to when add_attribute is used: 
        object.__setattr__(self,'_attributes',[])

        # Output file handle
        object.__setattr__(self,'_out_file',None)

        # Initialize from attribute list provided
        if attributes:
            for attr in attributes:
                self.add_attribute(attr,None)


    def __setattr__(self,name,value):
        r"""
        Check that attribute exists before setting it.
        If not, raise an AttributeError.
        Exception: attributes starting with '_' are ok to set.
        """

        if (not name in self._attributes) and (name[0] != '_'):
            print("*** Unrecognized attribute: ",name)
            print("*** Perhaps a typo?")
            print("*** Add new attributes using add_attribute method")
            raise AttributeError("Unrecognized attribute: %s" % name)
        
        # attribute exists, ok to set:
        object.__setattr__(self,name,value)


    def __str__(self):
        r"""Returns string representation of this object"""
        output = "%s%s\n" % ("Name".ljust(25),"Value".ljust(12))
        for (k,v) in self.items():
            output += "%s%s\n" % (str(k).ljust(25),str(v).ljust(12))
        return output


    def add_attribute(self, name, value=None, add_to_list=True):
        r"""
        Adds an attribute called name to the data object

        If an attribute needs to be added to the object, this routine must be
        called or the attribute will not be written out.

        :Input:
         - *name* - (string) Name of the data attribute
         - *value* - (id) Value to set *name* to, defaults to None
        """
        if (name not in self._attributes) and add_to_list:
            self._attributes.append(name)
        object.__setattr__(self,name,value)


    def add_attributes(self, arg_list, value=None):
        r"""
        Add a list of attributes, each initialized to *value*.
        """
        for name in arg_list:
            self.add_attribute(name, value)


    def remove_attributes(self, arg_list):
        r"""
        Remove the listed attributes.
        """

        # Convert to list if args is not already a list
        if not isinstance(arg_list,list):
            arg_list = [arg_list]

        for arg in arg_list:
            self._attributes.remove(arg)
            delattr(self,arg)


    def attributes(self):
        r"""Returns tuple of attribute names"""
        return tuple(self._attributes)


    def has_attribute(self,name):
        r"""
        Check if this data object has the given attributes

        :Input:
         - *name* - (string) Name of attribute

        :Output:
         - (bool) - True if data object contains a data attribute name
        """
        return name in self._attributes

        
    def iteritems(self):
        r"""
        Returns an iterator of attributes and values from this object

        :Output:
         - (Iterator) Iterator over attributes and values
        """
        return [(k,getattr(self,k)) for k in self._attributes]


    def open_data_file(self, name, datasource='setrun.py'):
        """
        Open a data file and write a warning header.
        Warning header starts with '#' character.  These lines are skipped if
        data file is opened using the library routine opendatafile.

        :Input:
         - *name* - (string) Name of data file
         - *datasource* - (string) Source for the data

        :Output:
         - (file) - file object
        """

        source = datasource.ljust(25)
        self._out_file = open(name, 'w')
        self._out_file.write('########################################################\n')
        self._out_file.write('### DO NOT EDIT THIS FILE:  GENERATED AUTOMATICALLY ####\n')
        self._out_file.write('### To modify data, edit  %s ####\n' % source)
        self._out_file.write('###    and then "make .data"                        ####\n')
        self._out_file.write('########################################################\n\n')


    def close_data_file(self):
        r"""Close output data file"""
        self._out_file.close()
        self._out_file = None


    def write(self, out_file, data_source='setrun.py'):
        r"""Write out all data files in this ClawData object"""

        # Open data file
        self.open_data_file(out_file,data_source)

        # Write out list of attributes
        for (name,value) in self.iteritems():
            self.data_write(name)


    def data_write(self, name=None, value=None, alt_name=None, description=''):
        r"""
        Write out value to data file, in the form ::

           value =: name   # [description]

        Remove brackets and commas from lists, and replace booleans by T/F.

        :Input:
         - *name* - (string) normally a string defining the variable,
           ``if name==None``, write a blank line.
         - *description* - (string) optional description
        """
        if self._out_file is None:
            raise Exception("No file currently open for output.")

        # Defaults to the name of the variable requested
        if alt_name is None:
            alt_name = name

        if name is None and value is None:
            # Write out a blank line
            self._out_file.write('\n')
        else:
            # Use the value passed in instead of fetching from the data object
            if value is None:
                value = self.__getattribute__(name)

            # Convert value to an appropriate string repr
            if isinstance(value,np.ndarray):
                value = list(value)
            if isinstance(value,tuple) | isinstance(value,list):
                # Remove [], (), and ','
                string_value = repr(value)[1:-1]
                string_value = string_value.replace(',','')
            elif isinstance(value,bool):
                if value:
                    string_value = 'T'
                else:
                    string_value = 'F'
            else:
                string_value = repr(value)
            padded_value = string_value.ljust(20)
            padded_name = alt_name.ljust(20)
            if description != '':
                self._out_file.write('%s =: %s # %s \n' % 
                                        (padded_value, padded_name, description))
            else:
                self._out_file.write('%s =: %s\n' % 
                                    (padded_value, padded_name))
  

    def read(self,path,force=False):
        r"""Read and fill applicable data attributes.

        Note that if the data attribute is not found an exception will be
        raised unless the force argument is set to True in which case a new
        attribute will be added.
        """

        data_file = open(os.path.abspath(path),'r')

        for lineno,line in enumerate(data_file):
            if "=:" not in line:
                continue

            value, tail = line.split("=:")
            varname = tail.split()[0]

            # Set this parameter
            if self.has_attribute(varname) or force:
                value = self._parse_value(value)
                if not self.has_attribute(varname):
                    self.add_attribute(varname,value)
                else:
                    setattr(self,varname,value)
    

    def _parse_value(self,value):
        r"""
        Attempt to make sense of a value string from a config file.  If the
        value is not obviously an integer, float, or boolean, it is returned as
        a string stripped of leading and trailing whitespace.

        :Input:
            - *value* - (string) Value string to be parsed

        :Output:
            - (id) - Appropriate object based on *value*
        """
        value = value.strip()
        if not value:
            return None

        # assume that values containing spaces are lists of values
        if len(value.split()) > 1:
            return [self._parse_value(vv) for vv in value.split()]

        try:
            # see if it's an integer
            value = int(value)
        except ValueError:
            try:
                # see if it's a float
                value = float(value)
            except ValueError:
                # see if it's a bool
                if value[0] == 'T':
                    value = True
                elif value[0] == 'F':
                    value = False

        return value

#  Base data class for Clawpack data objects
# ==============================================================================


# ==============================================================================
# Clawpack input data classes
class ClawRunData(ClawData):
    r"""
    Object that contains all data objects that need to written out.

    Depending on the package type, this object contains the necessary data
    objects that need to eventually be written out to files.
    """

    def __init__(self, pkg, num_dim):
        super(ClawRunData,self).__init__()
        self.add_attribute('pkg',pkg)
        self.add_attribute('num_dim',num_dim)
        self.add_attribute('data_list',[])
        self.add_attribute('xclawcmd',None)

        # Always need the basic clawpack data object
        self.add_data(ClawInputData(num_dim),'clawdata')

        # Add package specific data objects
        if pkg.lower() in ['classic', 'classicclaw']:
            self.xclawcmd = 'xclaw'

        elif pkg.lower() in ['amrclaw', 'amr']:

            import clawpack.amrclaw.data as amrclaw

            self.xclawcmd = 'xamr'
            self.add_data(amrclaw.AmrclawInputData(self.clawdata),'amrdata')
            self.add_data(amrclaw.RegionData(num_dim=num_dim),'regiondata')
            self.add_data(amrclaw.FlagRegionData(num_dim=num_dim),'flagregiondata')
            self.add_data(amrclaw.GaugeData(num_dim=num_dim),'gaugedata')
            self.add_data(amrclaw.AdjointData(num_dim=num_dim),'adjointdata')

        elif pkg.lower() in ['geoclaw']:

            import clawpack.amrclaw.data as amrclaw
            import clawpack.geoclaw.data as geoclaw

            self.xclawcmd = 'xgeoclaw'

            # Required data set for basic run parameters:
            self.add_data(geoclaw.GeoClawData(),('geo_data'))
            self.add_data(amrclaw.GaugeData(num_dim=num_dim),'gaugedata')
            self.add_data(geoclaw.TopographyData(),'topo_data')
            self.add_data(geoclaw.DTopoData(),'dtopo_data')
            #self.add_data(geoclaw.BoussData(), 'bouss_data') # add to setrun

            if num_dim == 2:
                # options not available in 1d:
                self.add_data(amrclaw.AmrclawInputData(self.clawdata),'amrdata')
                self.add_data(amrclaw.AdjointData(num_dim=num_dim),
                              'adjointdata')
                self.add_data(amrclaw.RegionData(num_dim=num_dim),'regiondata')
                self.add_data(amrclaw.FlagRegionData(num_dim=num_dim),
                              'flagregiondata')
                self.add_data(geoclaw.RefinementData(),'refinement_data')
                self.add_data(geoclaw.FGoutData(),'fgout_data')
                self.add_data(geoclaw.FGmaxData(),'fgmax_data')
                self.add_data(geoclaw.QinitData(),'qinit_data')
                self.add_data(geoclaw.SurgeData(),'surge_data')
                self.add_data(geoclaw.FrictionData(),'friction_data')
                self.add_data(geoclaw.MultilayerData(), 'multilayer_data')
            elif num_dim == 1:
                self.add_data(geoclaw.GridData1D(), 'grid_data')
                #self.add_data(geoclaw.BoussData1D(), 'bouss_data')
                # explicitly add bouss_data in setrun when needed
            else:
                msg = 'Unexpected num_dim=%s for GeoClaw' % num_dim
                raise ValueError(msg)

        elif pkg.lower() in ['dclaw']:

            import clawpack.amrclaw.data as amrclaw
            import clawpack.geoclaw.data as geoclaw
            import clawpack.dclaw.data as dclaw

            self.xclawcmd = 'xgeoclaw'

            # Required data set for basic run parameters:
            self.add_data(geoclaw.GeoClawData(),('geo_data'))
            self.add_data(amrclaw.GaugeData(num_dim=num_dim),'gaugedata')
            self.add_data(geoclaw.TopographyData(),'topo_data')
            self.add_data(geoclaw.DTopoData(),'dtopo_data')

            if num_dim == 2:
                # options not available in 1d:
                self.add_data(amrclaw.AmrclawInputData(self.clawdata),'amrdata')
                # self.add_data(amrclaw.AdjointData(num_dim=num_dim),
                #               'adjointdata')
                self.add_data(amrclaw.RegionData(num_dim=num_dim),'regiondata')
                self.add_data(amrclaw.FlagRegionData(num_dim=num_dim),
                              'flagregiondata')
                self.add_data(geoclaw.RefinementData(),'refinement_data')
                self.add_data(geoclaw.FGoutData(),'fgout_data')
                self.add_data(geoclaw.FGmaxData(),'fgmax_data')
                # self.add_data(geoclaw.QinitData(),'qinit_data')
                # self.add_data(geoclaw.SurgeData(),'surge_data')
                # self.add_data(geoclaw.FrictionData(),'friction_data')
                # self.add_data(geoclaw.MultilayerData(), 'multilayer_data')

                self.add_data(dclaw.DClawInputData(),'dclaw_data')
                self.add_data(dclaw.QinitDClawData(),'qinitdclaw_data')
                self.add_data(dclaw.AuxInitDClawData(),'auxinitdclaw_data')
                self.add_data(dclaw.PInitDClawInputData(),'pinitdclaw_data')
                self.add_data(dclaw.FlowGradesData(),'flowgrades_data')

            else:
                msg = 'Unexpected num_dim=%s for DClaw' % num_dim
                raise ValueError(msg)
                

        else:
            raise AttributeError("Unrecognized Clawpack pkg = %s" % pkg)


    def add_data(self,data,name,file_name=None):
        r"""Add data object named *name* and written to *file_name*."""
        self.add_attribute(name,data)
        self.data_list.append(data)


    def replace_data(self, name, new_object):
        r"""
        Replace data objected named *name* with *new_object*
        """
        self.data_list.remove(getattr(self, name))
        setattr(self, name, new_object)
        self.data_list.append(getattr(self, name))


    def new_UserData(self,name,fname):
        r"""
        Create a new attribute called name
        for application specific data to be written
        to the data file fname.
        """
        data = UserData(fname)
        self.add_data(data,name,fname)
        return data


    def write(self, out_dir = ''):
        r"""Write out each data objects in datalist """
        
        import clawpack.amrclaw.data as amrclaw

        for data_object in self.data_list:
            # UserData doesn't naturally have an "out_file" parameter
            if isinstance(data_object, UserData):
                fname = data_object.__fname__
            else:
                argspec = inspect.signature(data_object.write)
                fname = argspec.parameters['out_file'].default
            fpath = os.path.join(out_dir,fname)
            if isinstance(data_object, amrclaw.GaugeData):
                data_object.write(self.clawdata.num_eqn, self.clawdata.num_aux, out_file=fpath)
            else:
                data_object.write(out_file=fpath)



class ClawInputData(ClawData):
    r"""
    Object containing basic Clawpack input data, usually written to 'claw.data'.


    """

    def __init__(self, num_dim):
        super(ClawInputData,self).__init__()

        # Set default values:
        self.add_attribute('num_dim',num_dim)
        self.add_attribute('num_eqn',1)
        self.add_attribute('num_waves',1)
        self.add_attribute('num_aux',0)
        self.add_attribute('output_style',1)
        self.add_attribute('output_times',[])
        self.add_attribute('num_output_times',None)
        self.add_attribute('output_t0',True)
        self.add_attribute('output_step_interval',None)
        self.add_attribute('total_steps',None)
        self.add_attribute('tfinal',None)
        self.add_attribute('output_format',1)
        self.add_attribute('output_q_components','all')
        self.add_attribute('output_aux_components','none')
        self.add_attribute('output_aux_onlyonce',True)
        
        self.add_attribute('dt_initial',1.e-5)
        self.add_attribute('dt_max',1.e99)
        self.add_attribute('dt_variable',True)
        self.add_attribute('cfl_desired',0.9)
        self.add_attribute('cfl_max',1.0)
        self.add_attribute('steps_max',50000)
        self.add_attribute('order',2)
        self.add_attribute('dimensional_split',0)
        self.add_attribute('verbosity',0)
        self.add_attribute('verbosity_regrid',0)
        self.add_attribute('source_split',0)
        self.add_attribute('capa_index',0)
        self.add_attribute('limiter',[4])
        self.add_attribute('t0',0.)
        self.add_attribute('num_ghost',2)
        self.add_attribute('use_fwaves',False)
        
        if num_dim == 1:
            self.add_attribute('lower',[0.])
            self.add_attribute('upper',[1.])
            self.add_attribute('num_cells',[100])
            self.add_attribute('bc_lower',[0])
            self.add_attribute('bc_upper',[0])
            self.add_attribute('transverse_waves',0)
        elif num_dim == 2:
            self.add_attribute('lower',[0.,0.])
            self.add_attribute('upper',[1.,1.])
            self.add_attribute('num_cells',[100,100])
            self.add_attribute('bc_lower',[0,0])
            self.add_attribute('bc_upper',[0,0])
            self.add_attribute('transverse_waves',2)
        elif num_dim == 3:
            self.add_attribute('lower',[0.,0.,0.])
            self.add_attribute('upper',[1.,1.,1.])
            self.add_attribute('num_cells',[100,100,100])
            self.add_attribute('bc_lower',[0,0,0])
            self.add_attribute('bc_upper',[0,0,0])
            self.add_attribute('transverse_waves',22)
        else:
            raise ValueError("Only num_dim=1, 2, or 3 supported ")

        # Restart capability, not all Clawpack packages handles this yet.
        self.add_attribute('restart',False)
        self.add_attribute('restart_file','')
        self.add_attribute('checkpt_style',0)
        self.add_attribute('checkpt_interval',1000)
        self.add_attribute('checkpt_time_interval',1000.)
        self.add_attribute('checkpt_times',[1000.])


    def write(self, out_file='claw.data', data_source='setrun.py'):
        r"""Write input data to a file"""
        self.open_data_file(out_file,data_source)

        self.data_write('num_dim')
        self.data_write('lower')
        self.data_write('upper')
        self.data_write('num_cells')
        self.data_write()  # writes blank line
        self.data_write('num_eqn')
        self.data_write('num_waves')
        self.data_write('num_aux')
        self.data_write()  # writes blank line

        self.data_write('t0')
        self.data_write()
        self.data_write('output_style')

        if self.output_style == 1:
            self.data_write('num_output_times')
            self.data_write('tfinal')
            self.data_write('output_t0')
        elif self.output_style == 2:
            if len(self.output_times) == 0:
                raise AttributeError("*** output_style==2 requires nonempty list" \
                        + " of output times")
            self.num_output_times = len(self.output_times)
            self.data_write('num_output_times')
            self.data_write('output_times')
        elif self.output_style==3:
            self.data_write('output_step_interval')
            self.data_write('total_steps')
            self.data_write('output_t0')
        else:
            raise AttributeError("*** Unrecognized output_style: %s"\
                  % self.output_style)

        self.data_write()
        if self.output_format in [1,'ascii']:
            self.output_format = 1
        elif self.output_format in [2,'binary32']:
            self.output_format = 2
        elif self.output_format in [3,'binary64','binary']:
            self.output_format = 3
        else:
            errmsg = "*** Error in data parameter: " + \
                     "output_format unrecognized: %s " % self.output_format + \
                     "\n     *** Expecting ascii, binary32, or binary64"
            raise ValueError(errmsg)
            
        self.data_write('output_format')

        if self.output_q_components == 'all':
            iout_q = self.num_eqn * [1]
        elif self.output_q_components == 'none':
            iout_q = self.num_eqn * [0]
        else:
            #iout_q = np.where(self.output_q_components, 1, 0)
            print("*** WARNING: Selective output_q_components not implemented")
            print("***          Will output all components of q")
            iout_q = self.num_eqn * [1]
    

        # Write out local value of iout_q rather than a data member
        self.data_write('', value=iout_q, alt_name='iout_q')

        if self.num_aux > 0:
            if isinstance(self.output_aux_components,str):
                if self.output_aux_components.lower() == 'all':
                    iout_aux = self.num_aux * [1]
                elif self.output_aux_components.lower() == 'none':
                    iout_aux = self.num_aux * [0]
                else:
                    raise ValueError("Invalid aux array component option.")
            else:
                iout_aux = np.where(self.output_aux_components, 1, 0)
                print("*** WARNING: Selective output_aux_components not implemented")
                print("***          Will output all components of aux")
                iout_aux = self.num_aux * [1]
            self.data_write(name='', value=iout_aux, alt_name='iout_aux')
            self.data_write('output_aux_onlyonce')

        self.data_write()
        self.data_write('dt_initial')
        self.data_write('dt_max')
        self.data_write('cfl_max')
        self.data_write('cfl_desired')
        self.data_write('steps_max')
        self.data_write()
        self.dt_variable = bool(self.dt_variable) # in case 0 or 1
        self.data_write('dt_variable')
        self.data_write('order')

        if self.num_dim == 1:
            pass
        else:
            # Transverse options different in 2D and 3D
            if self.num_dim == 2:
                if self.transverse_waves in [0,'none']:  
                    self.transverse_waves = 0
                elif self.transverse_waves in [1,'increment']:  
                    self.transverse_waves = 1
                elif self.transverse_waves in [2,'all']:  
                    self.transverse_waves = 2
                else:
                    raise AttributeError("Unrecognized transverse_waves: %s" \
                                             % self.transverse_waves)
            else:    # 3D
                if self.transverse_waves in [0,'none']:  
                    self.transverse_waves = 0
                elif self.transverse_waves in [1,'increment']:  
                    self.transverse_waves = 11
                elif self.transverse_waves in [2,'all']:  
                    self.transverse_waves = 22
                if not (self.transverse_waves in [0, 10, 11, 20, 21, 22]):
                    raise AttributeError("Unrecognized transverse_waves: %s" \
                                             % self.transverse_waves)
            self.data_write(None, self.transverse_waves, 'transverse_waves')

            if self.dimensional_split in [0,'unsplit']:  
                self.dimensional_split = 0
            elif self.dimensional_split in [1,'godunov']:  
                self.dimensional_split = 1
            elif self.dimensional_split in [2,'strang']:  
                if self.num_dim == 3:
                    raise AttributeError("Strang dimensional splitting not supported in 3D")
                else:
                    self.dimensional_split = 2
            else:
                raise AttributeError("Unrecognized dimensional_split: %s" \
                      % self.dimensional_split)
            self.data_write('dimensional_split')
            
        self.data_write('verbosity')

        if self.source_split in [0,'none']:  
            self.source_split = 0
        elif self.source_split in [1,'godunov']:  
            self.source_split = 1
        elif self.source_split in [2,'strang']:  
            self.source_split = 2
        else:
            raise AttributeError("Unrecognized source_split: %s" \
                  % self.source_split)
        self.data_write('source_split')

        self.data_write('capa_index')
        self.data_write('use_fwaves')
        self.data_write()

        for i in range(len(self.limiter)):
            if self.limiter[i] in [0,'none']:        self.limiter[i] = 0
            elif self.limiter[i] in [1,'minmod']:    self.limiter[i] = 1
            elif self.limiter[i] in [2,'superbee']:  self.limiter[i] = 2
            elif self.limiter[i] in [3,'vanleer']:   self.limiter[i] = 3
            elif self.limiter[i] in [4,'mc']:        self.limiter[i] = 4
            elif self.limiter[i] in [5,'bw']:        self.limiter[i] = 5
            else:
                raise AttributeError("Unrecognized limiter: %s" \
                      % self.limiter[i])
        self.data_write('limiter')

        self.data_write()

        self.data_write('num_ghost')
        if not isinstance(self.bc_lower, list):
            self.bc_lower = [self.bc_lower]    # Allow bare number in 1D
        if len(self.bc_lower) != self.num_dim:
            raise AttributeError("Incorrect number of lower BC codes (expected %d, got %d)" \
                                     %(self.num_dim, len(self.bc_lower)))
        for i in range(self.num_dim):
            if self.bc_lower[i] in [0,'user']:       self.bc_lower[i] = 0
            elif self.bc_lower[i] in [1,'extrap']:   self.bc_lower[i] = 1
            elif self.bc_lower[i] in [2,'periodic']: self.bc_lower[i] = 2
            elif self.bc_lower[i] in [3,'wall']:     self.bc_lower[i] = 3
            else:
                raise AttributeError("Unrecognized bc_lower: %s" \
                      % self.bc_lower[i])
        self.data_write('bc_lower')

        if not isinstance(self.bc_upper, list):
            self.bc_upper = [self.bc_upper]    # Allow bare number in 1D
        if len(self.bc_upper) != self.num_dim:
            raise AttributeError("Incorrect number of upper BC codes (expected %d, got %d)" \
                                     %(self.num_dim, len(self.bc_upper)))
        for i in range(self.num_dim):
            if self.bc_upper[i] in [0,'user']:       self.bc_upper[i] = 0
            elif self.bc_upper[i] in [1,'extrap']:   self.bc_upper[i] = 1
            elif self.bc_upper[i] in [2,'periodic']: self.bc_upper[i] = 2
            elif self.bc_upper[i] in [3,'wall']:     self.bc_upper[i] = 3
            else:
                raise AttributeError("Unrecognized bc_upper: %s" \
                      % self.bc_upper[i])
        self.data_write('bc_upper')

        self.data_write()
        self.data_write('restart')
        self.data_write('restart_file')
        self.data_write('checkpt_style')

        if self.checkpt_style in [-2,2]:
            num_checkpt_times = len(self.checkpt_times)
            self.data_write(name='', value=num_checkpt_times, alt_name='num_checkpt_times')
            self.data_write('checkpt_times')
        elif self.checkpt_style in [-3,3]:
            self.data_write('checkpt_interval')
        elif self.checkpt_style not in [0,1,-1,-4,4]:
            raise AttributeError("*** Unrecognized checkpt_style: %s"\
                  % self.checkpt_style)

        self.data_write()
        self.close_data_file()


class UserData(ClawData):
    r"""
    Object that will be written out to user file such as setprob.data, as
    determined by the fname attribute.
    """

    def __init__(self, fname):

        super(UserData,self).__init__()

        # Create attributes without adding to attributes list:

        # file to be read by Fortran for this data:
        object.__setattr__(self,'__fname__',fname)

        # dictionary to hold descriptions:
        object.__setattr__(self,'__descr__',{})

    def add_param(self,name,value,descr=''):
         self.add_attribute(name,value)
         descr_dict = self.__descr__
         descr_dict[name] = descr

    def write(self, data_source='setrun.py', out_file=None):
        if out_file is None:
            out_file = self.__fname__
        super(UserData,self).write(out_file, data_source)
        self.close_data_file()




""" 
Set up the plot figures, axes, and items to be done for each frame.

This module is imported by the plotting routines and then the
function setplot is called to set the plot parameters.
    
""" 

import os

#--------------------------
def setplot(plotdata):
#--------------------------
    
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of pyclaw.plotters.data.ClawPlotData.
    Output: a modified version of plotdata.
    
    """ 


    from clawpack.visclaw import colormaps

    plotdata.clearfigures()  # clear any old figures,axes,items data
    plotdata.format = 'binary'      # 'ascii', 'binary', 'netcdf'
    

    # Figure for pressure
    # -------------------

    plotfigure = plotdata.new_plotfigure(name='Adjoint', figno=1)
    plotfigure.kwargs = {'figsize': (5.5,5)}

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.axescmd = 'axes([0.1,0.1,0.8,0.8])'
    plotaxes.xlimits = [-4,8]
    plotaxes.ylimits = [-1,11]
    plotaxes.title = 'Adjoint'
    plotaxes.scaled = True      # so aspect ratio is 1
    plotaxes.afteraxes = fixup_adjoint

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = 0
    plotitem.pcolor_cmap = colormaps.blue_white_red
    plotitem.pcolor_cmin = -0.1
    plotitem.pcolor_cmax = 0.1
    plotitem.add_colorbar = False
    

    # Parameters used only when creating html and/or latex hardcopy
    # e.g., via pyclaw.plotters.frametools.printframes:

    plotdata.printfigs = True                # print figures
    plotdata.print_format = 'png'            # file format
    plotdata.print_framenos = 'all'          # list of frames to print
    plotdata.print_fignos = 'all'            # list of figures to print
    plotdata.html = True                     # create html files of plots?
    plotdata.html_homelink = '../README.html'   # pointer for top of index
    plotdata.latex = True                    # create latex file of plots?
    plotdata.latex_figsperline = 2           # layout of plots
    plotdata.latex_framesperline = 1         # layout of plots
    plotdata.latex_makepdf = False           # also run pdflatex?

    return plotdata

# To plot gauge locations on pcolor or contour plot, use this as
# an afteraxis function:

def addgauges(current_data):
    from clawpack.visclaw import gaugetools
    gaugetools.plot_gauge_locations(current_data.plotdata, \
                                    gaugenos='all', format_string='ko', add_labels=True)

def fixup_adjoint(current_data):
    import pylab
    addgauges(current_data)



#!/usr/bin/env python

r"""
Cleans up by deleting object files, executable, and _output directory.
"""

import os,sys,glob
import shutil

if __name__ == "__main__":

    examples_dir = os.path.abspath('./examples')
    print "Will remove all _output, _plots, and build directories from ",examples_dir
    ans = raw_input("Ok? ")
    if ans.lower() not in ['y','yes']:
        print "Aborting."
        sys.exit()

    os.chdir(examples_dir)
    exdirlist = []
    for (dirpath, subdirs, files) in os.walk('.'):
        currentdir = os.path.abspath(os.getcwd())
        os.chdir(os.path.abspath(dirpath))
        
        print 'In directory ',dirpath
        
        if os.path.isdir('_output'):
            shutil.rmtree('./_output')
        if os.path.isdir('_plots'):
            shutil.rmtree('./_plots')
        if os.path.isdir('build'):
            shutil.rmtree('./build')
            
        os.chdir(currentdir)




"""
Regression tests for swril 3D advection.
"""

from __future__ import absolute_import
import sys
import os
import unittest

import clawpack.amrclaw.test as test


class Advection3DSwirlTest(test.AMRClawRegressionTest):
    r"""Basic test for a 3D advection test case"""


    def runTest(self, save=False):

        # Write out data files
        self.load_rundata()
        self.write_rundata_objects()

        # Run code
        self.run_code()

        # Perform tests
        self.check_frame(save=save, frame_num=1, file_name="regression_data_test2.txt")
        self.check_frame(save=save, frame_num=2, file_name="regression_data_test3.txt")

        self.check_gauges(save=save, gauge_id=1)
        self.check_gauges(save=save, gauge_id=2)

        self.success = True



if __name__=="__main__":
    if len(sys.argv) > 1:
        if bool(sys.argv[1]):
            # Fake the setup and save out output
            test = Advection3DSwirlTest()
            try:
                test.setUp()
                test.runTest(save=True)
            finally:
                test.tearDown()
            sys.exit(0)
    unittest.main()